# Copyright 2024 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.compat import get_stdout_text_writer
from awscli.customizations.commands import BasicCommand
from awscli.customizations.exceptions import ParamValidationError
from awscli.utils import is_a_tty

DESCRIPTION = (
    "Starts a Live Tail streaming session for one or more log groups. "
    "A Live Tail session provides a near real-time streaming of log events "
    "as they are ingested into selected log groups. "
    "A session can go on for a maximum of 3 hours.\n\n"
    "By default, this command start a native tailing session "
    "where recent log events appear from the bottom, "
    "the log events are output as-is in Plain text.  You can run this command with --mode interactive, "
    "which starts an interactive tailing session. Interactive tailing provides the ability to "
    "highlight up to 5 terms in your logs. "
    "The severity codes are highlighted by default. The logs are output in JSON format if possible, "
    "but can be toggled to Plain text if desired. Interactive experience is disabled if --color is set to off, "
    "or if the output doesn't allow for color. \n\n"
    "You must have logs:StartLiveTail permission to perform this operation. "
    "If the log events matching the filters are more than 500 events per second, "
    "we sample the events to provide the real-time tailing experience.\n\n"
    "If you are using CloudWatch cross-account observability, "
    "you can use this operation in a monitoring account and start tailing on Log Group(s) "
    "present in the linked source accounts. "
    "For more information, see "
    "https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/CloudWatch-Unified-Cross-Account.html.\n\n"
    "Live Tail sessions incur charges by session usage time, per minute. "
    "For pricing details, please refer to "
    "https://aws.amazon.com/cloudwatch/pricing/."
)

LIST_SCHEMA = {"type": "array", "items": {"type": "string"}}

LOG_GROUP_IDENTIFIERS = {
    "name": "log-group-identifiers",
    "required": True,
    "positional_arg": False,
    "nargs": "+",
    "schema": LIST_SCHEMA,
    "help_text": (
        "The Log Group Identifiers are the ARNs for the CloudWatch Logs groups to tail. "
        "You can provide up to 10 Log Group Identifiers.\n\n"
        "Logs can be filtered by Log Stream(s) by providing  "
        "--log-stream-names or --log-stream-name-prefixes. "
        "If more than one Log Group is provided --log-stream-names and "
        "--log-stream-name-prefixes  is disabled. "
        "--log-stream-names and --log-stream-name-prefixes can't be provided simultaneously.\n\n"
        "Note -  The Log Group ARN must be in the following format. Replace REGION and "
        "ACCOUNT_ID with your Region and account ID. "
        "``arn:aws:logs:REGION :ACCOUNT_ID :log-group:LOG_GROUP_NAME``. A ``:*`` after the ARN is prohibited."
        "For more information about ARN format, "
        'see <a href="https://docs.aws.amazon.com/AmazonCloudWatch/latest/logs/iam-access-control-overview-cwl.html">CloudWatch Logs resources and operations</a>.'
    ),
}

LOG_STREAM_NAMES = {
    "name": "log-stream-names",
    "positional_arg": False,
    "nargs": "+",
    "schema": LIST_SCHEMA,
    "help_text": (
        "The list of stream names to filter logs by.\n\n This parameter cannot be "
        "specified when --log-stream-name-prefixes are also specified. "
        "This parameter cannot be specified when multiple log-group-identifiers are specified"
    ),
}

LOG_STREAM_NAME_PREFIXES = {
    "name": "log-stream-name-prefixes",
    "positional_arg": False,
    "nargs": "+",
    "schema": LIST_SCHEMA,
    "help_text": (
        "The prefix to filter logs by. Only events from log streams with names beginning "
        "with this prefix will be returned. \n\nThis parameter cannot be specified when "
        "--log-stream-names is also specified. This parameter cannot be specified when "
        "multiple log-group-identifiers are specified"
    ),
}

LOG_EVENT_FILTER_PATTERN = {
    "name": "log-event-filter-pattern",
    "positional_arg": False,
    "cli_type_name": "string",
    "help_text": (
        "The filter pattern to use. "
        'See <a href="https://docs.aws.amazon.com/AmazonCloudWatch/latest/logs/FilterAndPatternSyntax.html">Filter and Pattern Syntax</a> '
        "for details. If not provided, all the events are matched. "
        "This option can be used to include or exclude log events patterns.  "
        "Additionally, when multiple filter patterns are provided, they must be encapsulated by quotes."
    ),
}

MODE = {
    "name": "mode",
    "positional_arg": False,
    "default": "print-only",
    "choices": ["interactive", "print-only"],
    "help_text": (
        "The command support the following modes:\n\n"
        "<ul>"
        "<li> interative - "
        "Starts a Live Tail session in interactive mode. "
        "In an interactive session, you can highlight as many as five terms in the tailed logs. "
        "The severity codes are highlighted by default. Press h to enter the highlight "
        "mode and then type in the terms to be highlighted, "
        "one at a time, and press enter. Press c to clear the highlighted term(s). "
        "Press t to toggle formatting between JSON/Plain text. "
        "Press Esc to exit the Live Tail session. The interactive experience is disabled if you specify —color as off, "
        "or if coloring is not allowed by your output. "
        "Press up / down keys to scroll up or down between log events, "
        "use Ctrl + u / Ctrl + d to scroll faster. Press q to scroll to latest log events."
        "</li>"
        "<li> print-only - "
        "Starts a LiveTail session in print-only mode. "
        "In print-only mode logs are tailed and are printed as-is. This mode can be used "
        "in conjunction with other shell commands. "
        "</li>"
        "</ul>"
    ),
}


class ColorNotAllowedInInteractiveModeError(ParamValidationError):
    pass


class StartLiveTailCommand(BasicCommand):
    NAME = "start-live-tail"
    DESCRIPTION = DESCRIPTION
    ARG_TABLE = [
        LOG_GROUP_IDENTIFIERS,
        LOG_STREAM_NAMES,
        LOG_STREAM_NAME_PREFIXES,
        LOG_EVENT_FILTER_PATTERN,
        MODE,
    ]

    def __init__(self, session):
        super().__init__(session)
        self._output = get_stdout_text_writer()

    def _get_client(self, parsed_globals):
        return self._session.create_client(
            "logs",
            region_name=parsed_globals.region,
            endpoint_url=parsed_globals.endpoint_url,
            verify=parsed_globals.verify_ssl,
        )

    def _get_start_live_tail_kwargs(self, parsed_args):
        kwargs = {"logGroupIdentifiers": parsed_args.log_group_identifiers}

        if parsed_args.log_stream_names is not None:
            kwargs["logStreamNames"] = parsed_args.log_stream_names
        if parsed_args.log_stream_name_prefixes is not None:
            kwargs["logStreamNamePrefixes"] = (
                parsed_args.log_stream_name_prefixes
            )
        if parsed_args.log_event_filter_pattern is not None:
            kwargs["logEventFilterPattern"] = (
                parsed_args.log_event_filter_pattern
            )

        return kwargs

    def _is_color_allowed(self, color):
        if color == "on":
            return True
        elif color == "off":
            return False
        return is_a_tty()

    def _run_main(self, parsed_args, parsed_globals):
        from awscli.customizations.logs.ui import (
            InteractiveUI,
            LiveTailLogEventsCollector,
            LiveTailSessionMetadata,
            PrintOnlyUI,
        )

        self._client = self._get_client(parsed_globals)

        start_live_tail_kwargs = self._get_start_live_tail_kwargs(parsed_args)
        response = self._client.start_live_tail(**start_live_tail_kwargs)

        log_events = []
        session_metadata = LiveTailSessionMetadata()

        ui = PrintOnlyUI(self._output, log_events)
        if parsed_args.mode == "interactive":
            if not self._is_color_allowed(parsed_globals.color):
                raise ColorNotAllowedInInteractiveModeError(
                    "Color is not allowed by your output. Use print-only mode or enable color."
                )
            ui = InteractiveUI(log_events, session_metadata)

        log_events_collector = LiveTailLogEventsCollector(
            self._output,
            ui,
            response["responseStream"],
            log_events,
            session_metadata,
        )
        log_events_collector.daemon = True

        log_events_collector.start()
        ui.run()

        log_events_collector.stop()



# Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import os
import pathlib
import site
import subprocess
import sys

from constants import (
    BIN_DIRNAME,
    CLI_SCRIPTS,
    DISTRIBUTION_SOURCE_SANDBOX,
    DOWNLOAD_DEPS_BOOTSTRAP_LOCK,
    IS_WINDOWS,
    PORTABLE_EXE_REQUIREMENTS_LOCK,
    PYTHON_EXE_NAME,
    ROOT_DIR,
    SYSTEM_SANDBOX_REQUIREMENTS_LOCK,
    ArtifactType,
)
from utils import Utils


class AwsCliVenv:
    _PARENT_SCRIPTS_TO_COPY = [
        "pyinstaller",
        "pyinstaller.exe",
    ]

    def __init__(self, venv_dir: str, utils: Utils = None):
        self._venv_dir = venv_dir
        if utils is None:
            utils = Utils()
        self._utils = utils

    def create(self):
        self._utils.create_venv(self._venv_dir, with_pip=True)

    def bootstrap(
        self, artifact_type: ArtifactType, download_deps: bool = False
    ):
        if download_deps:
            self._install_requirements(DOWNLOAD_DEPS_BOOTSTRAP_LOCK)
            if artifact_type == ArtifactType.PORTABLE_EXE.value:
                self._install_requirements(PORTABLE_EXE_REQUIREMENTS_LOCK)
            else:
                self._install_requirements(SYSTEM_SANDBOX_REQUIREMENTS_LOCK)
        else:
            self._copy_parent_packages()
        self._install_awscli()
        self._update_metadata()
        self._update_windows_script_header()

    def _copy_parent_packages(self):
        for site_package in site.getsitepackages():
            self._utils.copy_directory_contents_into(
                site_package, self._site_packages()
            )
        parent_scripts = pathlib.Path(sys.executable).parents[0]
        for script in self._PARENT_SCRIPTS_TO_COPY:
            source = os.path.join(parent_scripts, script)
            if self._utils.path_exists(source):
                self._utils.copy_file(
                    source, os.path.join(self.bin_dir, script)
                )

    def _install_requirements(self, requirements_file, cwd=None):
        self._pip_install(
            ["--no-build-isolation", "-r", requirements_file],
            cwd=cwd,
        )

    def _install_awscli(self):
        self._pip_install(
            [
                ROOT_DIR,
                "--no-build-isolation",
                "--no-cache-dir",
                "--no-index",
            ]
        )

    def _update_windows_script_header(self):
        # When installing to a venv pip will rewrite shebang lines
        # to reference the relevant virutalenv directly. This is not
        # the case for aws.cmd which is our own windows cmd file
        # and does not have a shebang that is re-writable.
        # We need to manually overwrite the header line in this script
        # to reference the current virtualenv.
        # If we are not on Windows then this is not relevant.
        if not IS_WINDOWS:
            return
        python_exe_path = os.path.join(self.bin_dir, "python.exe")
        exe_path = os.path.join(self.bin_dir, "aws.cmd")
        lines = self._utils.read_file_lines(exe_path)
        lines[0] = self._utils.get_script_header(python_exe_path)
        self._utils.write_file(exe_path, "".join(lines))

    def _update_metadata(self):
        self._utils.update_metadata(
            self._site_packages(),
            distribution_source=DISTRIBUTION_SOURCE_SANDBOX,
        )

    @property
    def bin_dir(self):
        return os.path.join(self._venv_dir, BIN_DIRNAME)

    @property
    def python_exe(self):
        return os.path.join(self.bin_dir, PYTHON_EXE_NAME)

    def _pip_install(self, args, cwd=None):
        args = [self.python_exe, "-m", "pip", "install"] + args
        run_kwargs = {"check": True}
        if IS_WINDOWS:
            args = " ".join([str(a) for a in args])
            # The tests on windows will fail when executed with
            # the wrapper test runner script in scripts/ci if this
            # is not executed from shell.
            run_kwargs["shell"] = True
        if cwd is not None:
            run_kwargs["cwd"] = cwd
        self._utils.run(args, **run_kwargs)

    def _site_packages(self) -> str:
        # On windows the getsitepackages can return the root venv dir.
        # So instead of just taking the first entry, we need to take the
        # first entry that contains the string "site-packages" in the path.
        site_path = [
            path
            for path in json.loads(
                subprocess.check_output(
                    [
                        self.python_exe,
                        "-c",
                        "import site, json; print(json.dumps(site.getsitepackages()))",
                    ]
                )
                .decode()
                .strip()
            )
            if "site-packages" in path
        ][0]
        return site_path



import os
import random
import string
import tempfile
import zipfile
from contextlib import closing, contextmanager

import botocore.session
import pytest
from botocore.stub import Stubber

from awscli.customizations.cloudformation import exceptions
from awscli.customizations.cloudformation.artifact_exporter import (
    GLOBAL_EXPORT_DICT,
    ApiGatewayRestApiResource,
    AppSyncFunctionConfigurationRequestTemplateResource,
    AppSyncFunctionConfigurationResponseTemplateResource,
    AppSyncResolverRequestTemplateResource,
    AppSyncResolverResponseTemplateResource,
    CloudFormationStackResource,
    CodeCommitRepositoryS3Resource,
    ElasticBeanstalkApplicationVersion,
    GlueJobCommandScriptLocationResource,
    GraphQLSchemaResource,
    LambdaFunctionResource,
    LambdaLayerVersionResource,
    Resource,
    ResourceWithS3UrlDict,
    ServerlessApiResource,
    ServerlessApplicationResource,
    ServerlessFunctionResource,
    ServerlessLayerVersionResource,
    ServerlessRepoApplicationLicense,
    ServerlessRepoApplicationReadme,
    ServerlessStateMachineDefinitionResource,
    StepFunctionsStateMachineDefinitionResource,
    Template,
    copy_to_temp_dir,
    include_transform_export_handler,
    is_local_file,
    is_local_folder,
    is_s3_url,
    make_abs_path,
    make_zip,
    parse_s3_url,
    upload_local_artifacts,
    zip_folder,
)
from awscli.testutils import FileCreator, mock
from tests.unit.customizations.cloudformation import BaseYAMLTest

VALID_CASES = [
    "s3://foo/bar",
    "s3://foo/bar/baz/cat/dog",
    "s3://foo/bar?versionId=abc",
    "s3://foo/bar/baz?versionId=abc&versionId=123",
    "s3://foo/bar/baz?versionId=abc",
    "s3://www.amazon.com/foo/bar",
    "s3://my-new-bucket/foo/bar?a=1&a=2&a=3&b=1",
]

INVALID_CASES = [
    # For purposes of exporter, we need S3 URLs to point to an object
    # and not a bucket
    "s3://foo",
    # two versionIds is invalid
    "https://s3-eu-west-1.amazonaws.com/bucket/key",
    "https://www.amazon.com",
]


@pytest.mark.parametrize("url", VALID_CASES)
def test_is_valid_s3_url(url):
    assert is_s3_url(url), f"{url} should be valid"


@pytest.mark.parametrize("url", INVALID_CASES)
def test_is_invalid_s3_url(url):
    assert not is_s3_url(url), f"{url} should be invalid"


UPLOADED_S3_URL = "s3://foo/bar?versionId=baz"

RESOURCE_EXPORT_TEST_CASES = [
    {"class": ServerlessFunctionResource, "expected_result": UPLOADED_S3_URL},
    {"class": ServerlessApiResource, "expected_result": UPLOADED_S3_URL},
    {"class": GraphQLSchemaResource, "expected_result": UPLOADED_S3_URL},
    {
        "class": AppSyncResolverRequestTemplateResource,
        "expected_result": UPLOADED_S3_URL,
    },
    {
        "class": AppSyncResolverResponseTemplateResource,
        "expected_result": UPLOADED_S3_URL,
    },
    {
        "class": AppSyncFunctionConfigurationRequestTemplateResource,
        "expected_result": UPLOADED_S3_URL,
    },
    {
        "class": AppSyncFunctionConfigurationResponseTemplateResource,
        "expected_result": UPLOADED_S3_URL,
    },
    {
        "class": ApiGatewayRestApiResource,
        "expected_result": {"Bucket": "foo", "Key": "bar", "Version": "baz"},
    },
    {
        "class": LambdaFunctionResource,
        "expected_result": {
            "S3Bucket": "foo",
            "S3Key": "bar",
            "S3ObjectVersion": "baz",
        },
    },
    {
        "class": ElasticBeanstalkApplicationVersion,
        "expected_result": {"S3Bucket": "foo", "S3Key": "bar"},
    },
    {
        "class": LambdaLayerVersionResource,
        "expected_result": {
            "S3Bucket": "foo",
            "S3Key": "bar",
            "S3ObjectVersion": "baz",
        },
    },
    {
        "class": ServerlessLayerVersionResource,
        "expected_result": UPLOADED_S3_URL,
    },
    {
        "class": ServerlessRepoApplicationReadme,
        "expected_result": UPLOADED_S3_URL,
    },
    {
        "class": ServerlessRepoApplicationLicense,
        "expected_result": UPLOADED_S3_URL,
    },
    {
        "class": ServerlessRepoApplicationLicense,
        "expected_result": UPLOADED_S3_URL,
    },
    {
        "class": GlueJobCommandScriptLocationResource,
        "expected_result": {"ScriptLocation": UPLOADED_S3_URL},
    },
    {
        "class": StepFunctionsStateMachineDefinitionResource,
        "expected_result": {"Bucket": "foo", "Key": "bar", "Version": "baz"},
    },
    {
        "class": ServerlessStateMachineDefinitionResource,
        "expected_result": {"Bucket": "foo", "Key": "bar", "Version": "baz"},
    },
]


@pytest.mark.parametrize("test", RESOURCE_EXPORT_TEST_CASES)
def test_all_resources_export(test):
    mock_path = "awscli.customizations.cloudformation.artifact_exporter.upload_local_artifacts"
    with mock.patch(mock_path) as upload_local_artifacts_mock:
        _helper_verify_export_resources(
            test["class"], upload_local_artifacts_mock, test["expected_result"]
        )


def _helper_verify_export_resources(
    test_class, upload_local_artifacts_mock, expected_result
):
    s3_uploader_mock = mock.Mock()
    upload_local_artifacts_mock.reset_mock()

    resource_id = "id"
    parent_dir = "dir"

    if '.' in test_class.PROPERTY_NAME:
        reversed_property_names = test_class.PROPERTY_NAME.split('.')
        reversed_property_names.reverse()
        property_dict = {reversed_property_names[0]: "foo"}
        for sub_property_name in reversed_property_names[1:]:
            property_dict = {sub_property_name: property_dict}
        resource_dict = property_dict
    else:
        resource_dict = {test_class.PROPERTY_NAME: "foo"}

    upload_local_artifacts_mock.return_value = UPLOADED_S3_URL
    resource_obj = test_class(s3_uploader_mock)
    resource_obj.export(resource_id, resource_dict, parent_dir)

    upload_local_artifacts_mock.assert_called_once_with(
        resource_id,
        resource_dict,
        test_class.PROPERTY_NAME,
        parent_dir,
        s3_uploader_mock,
    )
    if '.' in test_class.PROPERTY_NAME:
        top_level_property_name = test_class.PROPERTY_NAME.split('.')[0]
        result = resource_dict[top_level_property_name]
    else:
        result = resource_dict[test_class.PROPERTY_NAME]

    assert result == expected_result


class TestArtifactExporter(BaseYAMLTest):
    def setUp(self):
        super(TestArtifactExporter, self).setUp()
        self.s3_uploader_mock = mock.Mock()
        self.s3_uploader_mock.s3.meta.endpoint_url = (
            "https://s3.some-valid-region.amazonaws.com"
        )

    def test_parse_s3_url(self):
        valid = [
            {"url": "s3://foo/bar", "result": {"Bucket": "foo", "Key": "bar"}},
            {
                "url": "s3://foo/bar/cat/dog",
                "result": {"Bucket": "foo", "Key": "bar/cat/dog"},
            },
            {
                "url": "s3://foo/bar/baz?versionId=abc&param1=val1&param2=val2",
                "result": {
                    "Bucket": "foo",
                    "Key": "bar/baz",
                    "VersionId": "abc",
                },
            },
            {
                # VersionId is not returned if there are more than one versionId
                # keys in query parameter
                "url": "s3://foo/bar/baz?versionId=abc&versionId=123",
                "result": {"Bucket": "foo", "Key": "bar/baz"},
            },
        ]

        invalid = [
            # For purposes of exporter, we need S3 URLs to point to an object
            # and not a bucket
            "s3://foo",
            # two versionIds is invalid
            "https://s3-eu-west-1.amazonaws.com/bucket/key",
            "https://www.amazon.com",
        ]

        for config in valid:
            result = parse_s3_url(
                config["url"],
                bucket_name_property="Bucket",
                object_key_property="Key",
                version_property="VersionId",
            )

            self.assertEqual(result, config["result"])

        for url in invalid:
            with self.assertRaises(ValueError):
                parse_s3_url(url)

    def test_is_local_file(self):
        with tempfile.NamedTemporaryFile() as handle:
            self.assertTrue(is_local_file(handle.name))
            self.assertFalse(is_local_folder(handle.name))

    def test_is_local_folder(self):
        with self.make_temp_dir() as filename:
            self.assertTrue(is_local_folder(filename))
            self.assertFalse(is_local_file(filename))

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.zip_and_upload"
    )
    def test_upload_local_artifacts_local_file(self, zip_and_upload_mock):
        # Case 1: Artifact path is a relative path
        # Verifies that we package local artifacts appropriately
        property_name = "property"
        resource_id = "resource_id"
        expected_s3_url = "s3://foo/bar?versionId=baz"

        self.s3_uploader_mock.upload_with_dedup.return_value = expected_s3_url

        with tempfile.NamedTemporaryFile() as handle:
            # Artifact is a file in the temporary directory
            artifact_path = handle.name
            parent_dir = tempfile.gettempdir()

            resource_dict = {property_name: artifact_path}
            result = upload_local_artifacts(
                resource_id,
                resource_dict,
                property_name,
                parent_dir,
                self.s3_uploader_mock,
            )
            self.assertEqual(result, expected_s3_url)

            # Internally the method would convert relative paths to absolute
            # path, with respect to the parent directory
            absolute_artifact_path = make_abs_path(parent_dir, artifact_path)
            self.s3_uploader_mock.upload_with_dedup.assert_called_with(
                absolute_artifact_path
            )

            zip_and_upload_mock.assert_not_called()

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.zip_and_upload"
    )
    def test_upload_local_artifacts_local_file_abs_path(
        self, zip_and_upload_mock
    ):
        # Case 2: Artifact path is an absolute path
        # Verifies that we package local artifacts appropriately
        property_name = "property"
        resource_id = "resource_id"
        expected_s3_url = "s3://foo/bar?versionId=baz"

        self.s3_uploader_mock.upload_with_dedup.return_value = expected_s3_url

        with tempfile.NamedTemporaryFile() as handle:
            parent_dir = tempfile.gettempdir()
            artifact_path = make_abs_path(parent_dir, handle.name)

            resource_dict = {property_name: artifact_path}
            result = upload_local_artifacts(
                resource_id,
                resource_dict,
                property_name,
                parent_dir,
                self.s3_uploader_mock,
            )
            self.assertEqual(result, expected_s3_url)

            self.s3_uploader_mock.upload_with_dedup.assert_called_with(
                artifact_path
            )
            zip_and_upload_mock.assert_not_called()

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.zip_and_upload"
    )
    def test_upload_local_artifacts_local_folder(self, zip_and_upload_mock):
        property_name = "property"
        resource_id = "resource_id"
        expected_s3_url = "s3://foo/bar?versionId=baz"

        zip_and_upload_mock.return_value = expected_s3_url

        #  Artifact path is a Directory
        with self.make_temp_dir() as artifact_path:
            # Artifact is a file in the temporary directory
            parent_dir = tempfile.gettempdir()
            resource_dict = {property_name: artifact_path}

            result = upload_local_artifacts(
                resource_id,
                resource_dict,
                property_name,
                parent_dir,
                mock.Mock(),
            )
            self.assertEqual(result, expected_s3_url)

            absolute_artifact_path = make_abs_path(parent_dir, artifact_path)

            zip_and_upload_mock.assert_called_once_with(
                absolute_artifact_path, mock.ANY
            )

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.zip_and_upload"
    )
    def test_upload_local_artifacts_no_path(self, zip_and_upload_mock):
        property_name = "property"
        resource_id = "resource_id"
        expected_s3_url = "s3://foo/bar?versionId=baz"

        zip_and_upload_mock.return_value = expected_s3_url

        # If you don't specify a path, we will default to Current Working Dir
        resource_dict = {}
        parent_dir = tempfile.gettempdir()

        result = upload_local_artifacts(
            resource_id,
            resource_dict,
            property_name,
            parent_dir,
            self.s3_uploader_mock,
        )
        self.assertEqual(result, expected_s3_url)

        zip_and_upload_mock.assert_called_once_with(parent_dir, mock.ANY)
        self.s3_uploader_mock.upload_with_dedup.assert_not_called()

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.zip_and_upload"
    )
    def test_upload_local_artifacts_s3_url(self, zip_and_upload_mock):
        property_name = "property"
        resource_id = "resource_id"
        object_s3_url = "s3://foo/bar?versionId=baz"

        # If URL is already S3 URL, this will be returned without zip/upload
        resource_dict = {property_name: object_s3_url}
        parent_dir = tempfile.gettempdir()

        result = upload_local_artifacts(
            resource_id,
            resource_dict,
            property_name,
            parent_dir,
            self.s3_uploader_mock,
        )
        self.assertEqual(result, object_s3_url)

        zip_and_upload_mock.assert_not_called()
        self.s3_uploader_mock.upload_with_dedup.assert_not_called()

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.zip_and_upload"
    )
    def test_upload_local_artifacts_invalid_value(self, zip_and_upload_mock):
        property_name = "property"
        resource_id = "resource_id"
        parent_dir = tempfile.gettempdir()

        with self.assertRaises(exceptions.InvalidLocalPathError):
            non_existent_file = "some_random_filename"
            resource_dict = {property_name: non_existent_file}
            upload_local_artifacts(
                resource_id,
                resource_dict,
                property_name,
                parent_dir,
                self.s3_uploader_mock,
            )

        with self.assertRaises(exceptions.InvalidLocalPathError):
            non_existent_file = ["invalid datatype"]
            resource_dict = {property_name: non_existent_file}
            upload_local_artifacts(
                resource_id,
                resource_dict,
                property_name,
                parent_dir,
                self.s3_uploader_mock,
            )

        zip_and_upload_mock.assert_not_called()
        self.s3_uploader_mock.upload_with_dedup.assert_not_called()

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.make_zip"
    )
    def test_zip_folder(self, make_zip_mock):
        zip_file_name = "name.zip"
        make_zip_mock.return_value = zip_file_name

        with self.make_temp_dir() as dirname:
            with zip_folder(dirname) as actual_zip_file_name:
                self.assertEqual(actual_zip_file_name, zip_file_name)

        make_zip_mock.assert_called_once_with(mock.ANY, dirname)

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.upload_local_artifacts"
    )
    def test_resource(self, upload_local_artifacts_mock):
        # Property value is a path to file

        class MockResource(Resource):
            PROPERTY_NAME = "foo"

        resource = MockResource(self.s3_uploader_mock)

        resource_id = "id"
        resource_dict = {}
        resource_dict[resource.PROPERTY_NAME] = "/path/to/file"
        parent_dir = "dir"
        s3_url = "s3://foo/bar"

        upload_local_artifacts_mock.return_value = s3_url

        resource.export(resource_id, resource_dict, parent_dir)

        upload_local_artifacts_mock.assert_called_once_with(
            resource_id,
            resource_dict,
            resource.PROPERTY_NAME,
            parent_dir,
            self.s3_uploader_mock,
        )

        self.assertEqual(resource_dict[resource.PROPERTY_NAME], s3_url)

    @mock.patch("shutil.rmtree")
    @mock.patch("zipfile.is_zipfile")
    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.copy_to_temp_dir"
    )
    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.zip_and_upload"
    )
    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.is_local_file"
    )
    def test_resource_with_force_zip_on_regular_file(
        self,
        is_local_file_mock,
        zip_and_upload_mock,
        copy_to_temp_dir_mock,
        is_zipfile_mock,
        rmtree_mock,
    ):
        # Property value is a path to file and FORCE_ZIP is True

        class MockResource(Resource):
            PROPERTY_NAME = "foo"
            FORCE_ZIP = True

        resource = MockResource(self.s3_uploader_mock)

        resource_id = "id"
        resource_dict = {}
        original_path = "/path/to/file"
        resource_dict[resource.PROPERTY_NAME] = original_path
        parent_dir = "dir"
        s3_url = "s3://foo/bar"

        zip_and_upload_mock.return_value = s3_url
        is_local_file_mock.return_value = True

        with self.make_temp_dir() as tmp_dir:
            copy_to_temp_dir_mock.return_value = tmp_dir

            # This is not a zip file
            is_zipfile_mock.return_value = False

            resource.export(resource_id, resource_dict, parent_dir)

            zip_and_upload_mock.assert_called_once_with(tmp_dir, mock.ANY)
            rmtree_mock.assert_called_once_with(tmp_dir)
            is_zipfile_mock.assert_called_once_with(original_path)
            assert resource_dict[resource.PROPERTY_NAME] == s3_url

    @mock.patch("shutil.rmtree")
    @mock.patch("zipfile.is_zipfile")
    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.copy_to_temp_dir"
    )
    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.zip_and_upload"
    )
    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.is_local_file"
    )
    def test_resource_with_force_zip_on_zip_file(
        self,
        is_local_file_mock,
        zip_and_upload_mock,
        copy_to_temp_dir_mock,
        is_zipfile_mock,
        rmtree_mock,
    ):
        # Property value is a path to zip file and FORCE_ZIP is True
        # We should *not* re-zip an existing zip

        class MockResource(Resource):
            PROPERTY_NAME = "foo"
            FORCE_ZIP = True

        resource = MockResource(self.s3_uploader_mock)

        resource_id = "id"
        resource_dict = {}
        original_path = "/path/to/zip_file"
        resource_dict[resource.PROPERTY_NAME] = original_path
        parent_dir = "dir"
        s3_url = "s3://foo/bar"

        # When the file is actually a zip-file, no additional zipping has to happen
        is_zipfile_mock.return_value = True
        is_local_file_mock.return_value = True
        zip_and_upload_mock.return_value = s3_url
        self.s3_uploader_mock.upload_with_dedup.return_value = s3_url

        resource.export(resource_id, resource_dict, parent_dir)

        copy_to_temp_dir_mock.assert_not_called()
        zip_and_upload_mock.assert_not_called()
        rmtree_mock.assert_not_called()
        is_zipfile_mock.assert_called_once_with(original_path)
        assert resource_dict[resource.PROPERTY_NAME] == s3_url

    @mock.patch("shutil.rmtree")
    @mock.patch("zipfile.is_zipfile")
    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.copy_to_temp_dir"
    )
    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.zip_and_upload"
    )
    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.is_local_file"
    )
    def test_resource_without_force_zip(
        self,
        is_local_file_mock,
        zip_and_upload_mock,
        copy_to_temp_dir_mock,
        is_zipfile_mock,
        rmtree_mock,
    ):
        class MockResourceNoForceZip(Resource):
            PROPERTY_NAME = "foo"

        resource = MockResourceNoForceZip(self.s3_uploader_mock)

        resource_id = "id"
        resource_dict = {}
        original_path = "/path/to/file"
        resource_dict[resource.PROPERTY_NAME] = original_path
        parent_dir = "dir"
        s3_url = "s3://foo/bar"

        # This is not a zip file, but a valid local file. Since FORCE_ZIP is NOT set, this will not be zipped
        is_zipfile_mock.return_value = False
        is_local_file_mock.return_value = True
        zip_and_upload_mock.return_value = s3_url
        self.s3_uploader_mock.upload_with_dedup.return_value = s3_url

        resource.export(resource_id, resource_dict, parent_dir)

        copy_to_temp_dir_mock.assert_not_called()
        zip_and_upload_mock.assert_not_called()
        rmtree_mock.assert_not_called()
        is_zipfile_mock.assert_called_once_with(original_path)
        assert resource_dict[resource.PROPERTY_NAME] == s3_url

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.upload_local_artifacts"
    )
    def test_resource_empty_property_value(self, upload_local_artifacts_mock):
        # Property value is empty

        class MockResource(Resource):
            PROPERTY_NAME = "foo"

        resource = MockResource(self.s3_uploader_mock)

        resource_id = "id"
        resource_dict = {}
        resource_dict[resource.PROPERTY_NAME] = "/path/to/file"
        parent_dir = "dir"
        s3_url = "s3://foo/bar"

        upload_local_artifacts_mock.return_value = s3_url
        resource_dict = {}
        resource.export(resource_id, resource_dict, parent_dir)
        upload_local_artifacts_mock.assert_called_once_with(
            resource_id,
            resource_dict,
            resource.PROPERTY_NAME,
            parent_dir,
            self.s3_uploader_mock,
        )
        self.assertEqual(resource_dict[resource.PROPERTY_NAME], s3_url)

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.upload_local_artifacts"
    )
    def test_resource_property_value_dict(self, upload_local_artifacts_mock):
        # Property value is a dictionary. Export should not upload anything

        class MockResource(Resource):
            PROPERTY_NAME = "foo"

        resource = MockResource(self.s3_uploader_mock)
        resource_id = "id"
        resource_dict = {}
        resource_dict[resource.PROPERTY_NAME] = "/path/to/file"
        parent_dir = "dir"
        s3_url = "s3://foo/bar"

        upload_local_artifacts_mock.return_value = s3_url
        resource_dict = {}
        resource_dict[resource.PROPERTY_NAME] = {"a": "b"}
        resource.export(resource_id, resource_dict, parent_dir)
        upload_local_artifacts_mock.assert_not_called()
        self.assertEqual(resource_dict, {"foo": {"a": "b"}})

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.upload_local_artifacts"
    )
    def test_resource_has_package_null_property_to_false(
        self, upload_local_artifacts_mock
    ):
        # Should not upload anything if PACKAGE_NULL_PROPERTY is set to False

        class MockResource(Resource):
            PROPERTY_NAME = "foo"
            PACKAGE_NULL_PROPERTY = False

        resource = MockResource(self.s3_uploader_mock)
        resource_id = "id"
        resource_dict = {}
        parent_dir = "dir"
        s3_url = "s3://foo/bar"

        upload_local_artifacts_mock.return_value = s3_url

        resource.export(resource_id, resource_dict, parent_dir)

        upload_local_artifacts_mock.assert_not_called()
        self.assertNotIn(resource.PROPERTY_NAME, resource_dict)

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.upload_local_artifacts"
    )
    def test_resource_export_fails(self, upload_local_artifacts_mock):
        class MockResource(Resource):
            PROPERTY_NAME = "foo"

        resource = MockResource(self.s3_uploader_mock)
        resource_id = "id"
        resource_dict = {}
        resource_dict[resource.PROPERTY_NAME] = "/path/to/file"
        parent_dir = "dir"
        s3_url = "s3://foo/bar"

        upload_local_artifacts_mock.side_effect = RuntimeError
        resource_dict = {}

        with self.assertRaises(exceptions.ExportFailedError):
            resource.export(resource_id, resource_dict, parent_dir)

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.upload_local_artifacts"
    )
    def test_resource_with_s3_url_dict(self, upload_local_artifacts_mock):
        """
        Checks if we properly export from the Resource classc
        :return:
        """

        self.assertTrue(issubclass(ResourceWithS3UrlDict, Resource))

        class MockResource(ResourceWithS3UrlDict):
            PROPERTY_NAME = "foo"
            BUCKET_NAME_PROPERTY = "b"
            OBJECT_KEY_PROPERTY = "o"
            VERSION_PROPERTY = "v"

        resource = MockResource(self.s3_uploader_mock)

        # Case 1: Property value is a path to file
        resource_id = "id"
        resource_dict = {}
        resource_dict[resource.PROPERTY_NAME] = "/path/to/file"
        parent_dir = "dir"
        s3_url = "s3://bucket/key1/key2?versionId=SomeVersionNumber"

        upload_local_artifacts_mock.return_value = s3_url

        resource.export(resource_id, resource_dict, parent_dir)

        upload_local_artifacts_mock.assert_called_once_with(
            resource_id,
            resource_dict,
            resource.PROPERTY_NAME,
            parent_dir,
            self.s3_uploader_mock,
        )

        self.assertEqual(
            resource_dict[resource.PROPERTY_NAME],
            {"b": "bucket", "o": "key1/key2", "v": "SomeVersionNumber"},
        )

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.Template"
    )
    def test_export_cloudformation_stack(self, TemplateMock):
        stack_resource = CloudFormationStackResource(self.s3_uploader_mock)

        resource_id = "id"
        property_name = stack_resource.PROPERTY_NAME
        exported_template_dict = {"foo": "bar"}
        result_s3_url = "s3://hello/world"
        result_path_style_s3_url = "http://s3.amazonws.com/hello/world"

        template_instance_mock = mock.Mock()
        TemplateMock.return_value = template_instance_mock
        template_instance_mock.export.return_value = exported_template_dict

        self.s3_uploader_mock.upload_with_dedup.return_value = result_s3_url
        self.s3_uploader_mock.to_path_style_s3_url.return_value = (
            result_path_style_s3_url
        )

        with tempfile.NamedTemporaryFile() as handle:
            template_path = handle.name
            resource_dict = {property_name: template_path}
            parent_dir = tempfile.gettempdir()

            stack_resource.export(resource_id, resource_dict, parent_dir)

            self.assertEqual(
                resource_dict[property_name], result_path_style_s3_url
            )

            TemplateMock.assert_called_once_with(
                template_path, parent_dir, self.s3_uploader_mock
            )
            template_instance_mock.export.assert_called_once_with()
            self.s3_uploader_mock.upload_with_dedup.assert_called_once_with(
                mock.ANY, "template"
            )
            self.s3_uploader_mock.to_path_style_s3_url.assert_called_once_with(
                "world", None
            )

    def _assert_stack_template_url_is_not_changed(self, s3_url):
        stack_resource = CloudFormationStackResource(self.s3_uploader_mock)
        resource_id = "id"
        property_name = stack_resource.PROPERTY_NAME
        resource_dict = {property_name: s3_url}

        # Case 1: Path is already S3 url
        stack_resource.export(resource_id, resource_dict, "dir")
        self.assertEqual(resource_dict[property_name], s3_url)
        self.s3_uploader_mock.upload_with_dedup.assert_not_called()

    def test_export_cloudformation_stack_no_upload_path_is_s3url(self):
        s3_url = "s3://hello/world"
        self._assert_stack_template_url_is_not_changed(s3_url)

    def test_export_cloudformation_stack_no_upload_path_is_httpsurl(self):
        s3_url = "https://s3.amazonaws.com/hello/world"
        self._assert_stack_template_url_is_not_changed(s3_url)

    def test_export_cloudformation_stack_no_upload_path_is_s3_region_httpsurl(
        self,
    ):
        s3_url = "https://s3.some-valid-region.amazonaws.com/hello/world"
        self._assert_stack_template_url_is_not_changed(s3_url)

    def test_export_cloudformation_stack_no_upload_path_is_virtual(self):
        s3_url = "https://bucket.s3.amazonaws.com/key"
        self._assert_stack_template_url_is_not_changed(s3_url)

    def test_export_cloudformation_stack_no_upload_path_is_virtual_region(
        self,
    ):
        s3_url = "https://bucket.s3.some-region.amazonaws.com/key"
        self._assert_stack_template_url_is_not_changed(s3_url)

    def test_export_cloudformation_stack_no_upload_path_is_empty(self):
        stack_resource = CloudFormationStackResource(self.s3_uploader_mock)
        resource_id = "id"
        property_name = stack_resource.PROPERTY_NAME
        s3_url = "s3://hello/world"
        resource_dict = {property_name: s3_url}

        # Case 2: Path is empty
        resource_dict = {}
        stack_resource.export(resource_id, resource_dict, "dir")
        self.assertEqual(resource_dict, {})
        self.s3_uploader_mock.upload_with_dedup.assert_not_called()

    def test_export_cloudformation_stack_no_upload_path_not_file(self):
        stack_resource = CloudFormationStackResource(self.s3_uploader_mock)
        resource_id = "id"
        property_name = stack_resource.PROPERTY_NAME
        s3_url = "s3://hello/world"

        # Case 3: Path is not a file
        with self.make_temp_dir() as dirname:
            resource_dict = {property_name: dirname}
            with self.assertRaises(exceptions.ExportFailedError):
                stack_resource.export(resource_id, resource_dict, "dir")
                self.s3_uploader_mock.upload_with_dedup.assert_not_called()

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.Template"
    )
    def test_export_serverless_application(self, TemplateMock):
        stack_resource = ServerlessApplicationResource(self.s3_uploader_mock)

        resource_id = "id"
        property_name = stack_resource.PROPERTY_NAME
        exported_template_dict = {"foo": "bar"}
        result_s3_url = "s3://hello/world"
        result_path_style_s3_url = "http://s3.amazonws.com/hello/world"

        template_instance_mock = mock.Mock()
        TemplateMock.return_value = template_instance_mock
        template_instance_mock.export.return_value = exported_template_dict

        self.s3_uploader_mock.upload_with_dedup.return_value = result_s3_url
        self.s3_uploader_mock.to_path_style_s3_url.return_value = (
            result_path_style_s3_url
        )

        with tempfile.NamedTemporaryFile() as handle:
            template_path = handle.name
            resource_dict = {property_name: template_path}
            parent_dir = tempfile.gettempdir()

            stack_resource.export(resource_id, resource_dict, parent_dir)

            self.assertEqual(
                resource_dict[property_name], result_path_style_s3_url
            )

            TemplateMock.assert_called_once_with(
                template_path, parent_dir, self.s3_uploader_mock
            )
            template_instance_mock.export.assert_called_once_with()
            self.s3_uploader_mock.upload_with_dedup.assert_called_once_with(
                mock.ANY, "template"
            )
            self.s3_uploader_mock.to_path_style_s3_url.assert_called_once_with(
                "world", None
            )

    def test_export_serverless_application_no_upload_path_is_s3url(self):
        stack_resource = ServerlessApplicationResource(self.s3_uploader_mock)
        resource_id = "id"
        property_name = stack_resource.PROPERTY_NAME
        s3_url = "s3://hello/world"
        resource_dict = {property_name: s3_url}

        # Case 1: Path is already S3 url
        stack_resource.export(resource_id, resource_dict, "dir")
        self.assertEqual(resource_dict[property_name], s3_url)
        self.s3_uploader_mock.upload_with_dedup.assert_not_called()

    def test_export_serverless_application_no_upload_path_is_httpsurl(self):
        stack_resource = ServerlessApplicationResource(self.s3_uploader_mock)
        resource_id = "id"
        property_name = stack_resource.PROPERTY_NAME
        s3_url = "https://s3.amazonaws.com/hello/world"
        resource_dict = {property_name: s3_url}

        # Case 1: Path is already S3 url
        stack_resource.export(resource_id, resource_dict, "dir")
        self.assertEqual(resource_dict[property_name], s3_url)
        self.s3_uploader_mock.upload_with_dedup.assert_not_called()

    def test_export_serverless_application_no_upload_path_is_empty(self):
        stack_resource = ServerlessApplicationResource(self.s3_uploader_mock)
        resource_id = "id"
        property_name = stack_resource.PROPERTY_NAME

        # Case 2: Path is empty
        resource_dict = {}
        stack_resource.export(resource_id, resource_dict, "dir")
        self.assertEqual(resource_dict, {})
        self.s3_uploader_mock.upload_with_dedup.assert_not_called()

    def test_export_serverless_application_no_upload_path_not_file(self):
        stack_resource = ServerlessApplicationResource(self.s3_uploader_mock)
        resource_id = "id"
        property_name = stack_resource.PROPERTY_NAME

        # Case 3: Path is not a file
        with self.make_temp_dir() as dirname:
            resource_dict = {property_name: dirname}
            with self.assertRaises(exceptions.ExportFailedError):
                stack_resource.export(resource_id, resource_dict, "dir")
                self.s3_uploader_mock.upload_with_dedup.assert_not_called()

    def test_export_serverless_application_no_upload_path_is_dictionary(self):
        stack_resource = ServerlessApplicationResource(self.s3_uploader_mock)
        resource_id = "id"
        property_name = stack_resource.PROPERTY_NAME

        # Case 4: Path is dictionary
        location = {"ApplicationId": "id", "SemanticVersion": "1.0.1"}
        resource_dict = {property_name: location}
        stack_resource.export(resource_id, resource_dict, "dir")
        self.assertEqual(resource_dict[property_name], location)
        self.s3_uploader_mock.upload_with_dedup.assert_not_called()

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.yaml_parse"
    )
    def test_template_export_metadata(self, yaml_parse_mock):
        parent_dir = os.path.abspath(os.path.sep)
        template_dir = os.path.join(parent_dir, 'foo', 'bar')
        template_path = os.path.join(template_dir, 'path')
        template_str = self.example_yaml_template()

        metadata_type1_class = mock.Mock()
        metadata_type1_class.RESOURCE_TYPE = "metadata_type1"
        metadata_type1_class.PROPERTY_NAME = "property_1"
        metadata_type1_instance = mock.Mock()
        metadata_type1_class.return_value = metadata_type1_instance

        metadata_type2_class = mock.Mock()
        metadata_type2_class.RESOURCE_TYPE = "metadata_type2"
        metadata_type2_class.PROPERTY_NAME = "property_2"
        metadata_type2_instance = mock.Mock()
        metadata_type2_class.return_value = metadata_type2_instance

        metadata_to_export = [metadata_type1_class, metadata_type2_class]

        template_dict = {
            "Metadata": {
                "metadata_type1": {"property_1": "abc"},
                "metadata_type2": {"property_2": "def"},
            }
        }
        open_mock = mock.mock_open()
        yaml_parse_mock.return_value = template_dict

        # Patch the file open method to return template string
        with mock.patch(
            "awscli.customizations.cloudformation."
            "artifact_exporter.compat_open",
            open_mock(read_data=template_str),
        ) as open_mock:
            template_exporter = Template(
                template_path,
                parent_dir,
                self.s3_uploader_mock,
                metadata_to_export=metadata_to_export,
            )
            exported_template = template_exporter.export()
            self.assertEqual(exported_template, template_dict)

            open_mock.assert_called_once_with(
                make_abs_path(parent_dir, template_path), "r"
            )

            self.assertEqual(1, yaml_parse_mock.call_count)

            metadata_type1_class.assert_called_once_with(self.s3_uploader_mock)
            metadata_type1_instance.export.assert_called_once_with(
                "metadata_type1", mock.ANY, template_dir
            )
            metadata_type2_class.assert_called_once_with(self.s3_uploader_mock)
            metadata_type2_instance.export.assert_called_once_with(
                "metadata_type2", mock.ANY, template_dir
            )

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.yaml_parse"
    )
    def test_template_export(self, yaml_parse_mock):
        parent_dir = os.path.abspath(os.path.sep)
        template_dir = os.path.join(parent_dir, 'foo', 'bar')
        template_path = os.path.join(template_dir, 'path')
        template_str = self.example_yaml_template()

        resource_type1_class = mock.Mock()
        resource_type1_class.RESOURCE_TYPE = "resource_type1"
        resource_type1_instance = mock.Mock()
        resource_type1_class.return_value = resource_type1_instance
        resource_type2_class = mock.Mock()
        resource_type2_class.RESOURCE_TYPE = "resource_type2"
        resource_type2_instance = mock.Mock()
        resource_type2_class.return_value = resource_type2_instance

        resources_to_export = [resource_type1_class, resource_type2_class]

        properties = {"foo": "bar"}
        template_dict = {
            "Resources": {
                "Resource1": {
                    "Type": "resource_type1",
                    "Properties": properties,
                },
                "Resource2": {
                    "Type": "resource_type2",
                    "Properties": properties,
                },
                "Resource3": {
                    "Type": "some-other-type",
                    "Properties": properties,
                },
            }
        }

        open_mock = mock.mock_open()
        yaml_parse_mock.return_value = template_dict

        # Patch the file open method to return template string
        with mock.patch(
            "awscli.customizations.cloudformation."
            "artifact_exporter.compat_open",
            open_mock(read_data=template_str),
        ) as open_mock:
            template_exporter = Template(
                template_path,
                parent_dir,
                self.s3_uploader_mock,
                resources_to_export,
            )
            exported_template = template_exporter.export()
            self.assertEqual(exported_template, template_dict)

            open_mock.assert_called_once_with(
                make_abs_path(parent_dir, template_path), "r"
            )

            self.assertEqual(1, yaml_parse_mock.call_count)

            resource_type1_class.assert_called_once_with(self.s3_uploader_mock)
            resource_type1_instance.export.assert_called_once_with(
                "Resource1", mock.ANY, template_dir
            )
            resource_type2_class.assert_called_once_with(self.s3_uploader_mock)
            resource_type2_instance.export.assert_called_once_with(
                "Resource2", mock.ANY, template_dir
            )

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.yaml_parse"
    )
    def test_template_export_foreach_valid(self, yaml_parse_mock):
        parent_dir = os.path.abspath(os.path.sep)
        template_dir = os.path.join(parent_dir, 'foo', 'bar')
        template_path = os.path.join(template_dir, 'path')
        template_str = self.example_yaml_template()

        resource_type1_class = mock.Mock()
        resource_type1_class.RESOURCE_TYPE = "resource_type1"
        resource_type1_instance = mock.Mock()
        resource_type1_class.return_value = resource_type1_instance
        resource_type2_class = mock.Mock()
        resource_type2_class.RESOURCE_TYPE = "resource_type2"
        resource_type2_instance = mock.Mock()
        resource_type2_class.return_value = resource_type2_instance

        resources_to_export = [resource_type1_class, resource_type2_class]

        properties = {"foo": "bar"}
        template_dict = {
            "Resources": {
                "Resource1": {
                    "Type": "resource_type1",
                    "Properties": properties,
                },
                "Resource2": {
                    "Type": "resource_type2",
                    "Properties": properties,
                },
                "Resource3": {
                    "Type": "some-other-type",
                    "Properties": properties,
                },
                "Fn::ForEach::OuterLoopName": [
                    "Identifier1",
                    ["4", "5"],
                    {
                        "Fn::ForEach::InnerLoopName": [
                            "Identifier2",
                            ["6", "7"],
                            {
                                "Resource${Identifier1}${Identifier2}": {
                                    "Type": "resource_type2",
                                    "Properties": properties,
                                }
                            },
                        ],
                        "Resource${Identifier1}": {
                            "Type": "resource_type1",
                            "Properties": properties,
                        },
                    },
                ],
            }
        }

        open_mock = mock.mock_open()
        yaml_parse_mock.return_value = template_dict

        # Patch the file open method to return template string
        with mock.patch(
            "awscli.customizations.cloudformation."
            "artifact_exporter.compat_open",
            open_mock(read_data=template_str),
        ) as open_mock:
            template_exporter = Template(
                template_path,
                parent_dir,
                self.s3_uploader_mock,
                resources_to_export,
            )
            exported_template = template_exporter.export()
            self.assertEqual(exported_template, template_dict)

            open_mock.assert_called_once_with(
                make_abs_path(parent_dir, template_path), "r"
            )

            self.assertEqual(1, yaml_parse_mock.call_count)

            resource_type1_class.assert_called_with(self.s3_uploader_mock)
            self.assertEqual(
                resource_type1_instance.export.call_args_list,
                [
                    mock.call("Resource1", properties, template_dir),
                    mock.call(
                        "Resource${Identifier1}", properties, template_dir
                    ),
                ],
            )
            resource_type2_class.assert_called_with(self.s3_uploader_mock)
            self.assertEqual(
                resource_type2_instance.export.call_args_list,
                [
                    mock.call("Resource2", properties, template_dir),
                    mock.call(
                        "Resource${Identifier1}${Identifier2}",
                        properties,
                        template_dir,
                    ),
                ],
            )

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.yaml_parse"
    )
    def test_template_export_foreach_invalid(self, yaml_parse_mock):
        parent_dir = os.path.abspath(os.path.sep)
        template_dir = os.path.join(parent_dir, 'foo', 'bar')
        template_path = os.path.join(template_dir, 'path')
        template_str = self.example_yaml_template()

        resource_type1_class = mock.Mock()
        resource_type1_class.RESOURCE_TYPE = "resource_type1"
        resource_type1_instance = mock.Mock()
        resource_type1_class.return_value = resource_type1_instance
        resource_type2_class = mock.Mock()
        resource_type2_class.RESOURCE_TYPE = "resource_type2"
        resource_type2_instance = mock.Mock()
        resource_type2_class.return_value = resource_type2_instance

        resources_to_export = [resource_type1_class, resource_type2_class]

        properties = {"foo": "bar"}
        template_dict = {
            "Resources": {
                "Resource1": {
                    "Type": "resource_type1",
                    "Properties": properties,
                },
                "Resource2": {
                    "Type": "resource_type2",
                    "Properties": properties,
                },
                "Resource3": {
                    "Type": "some-other-type",
                    "Properties": properties,
                },
                "Fn::ForEach::OuterLoopName": [
                    "Identifier1",
                    {"Resource${Identifier1}": {}},
                ],
            }
        }

        open_mock = mock.mock_open()
        yaml_parse_mock.return_value = template_dict

        # Patch the file open method to return template string
        with mock.patch(
            "awscli.customizations.cloudformation."
            "artifact_exporter.compat_open",
            open_mock(read_data=template_str),
        ) as open_mock:
            template_exporter = Template(
                template_path,
                parent_dir,
                self.s3_uploader_mock,
                resources_to_export,
            )
            with self.assertRaises(
                exceptions.InvalidForEachIntrinsicFunctionError
            ):
                template_exporter.export()

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.yaml_parse"
    )
    def test_template_global_export(self, yaml_parse_mock):
        parent_dir = os.path.abspath(os.path.sep)
        template_dir = os.path.join(parent_dir, 'foo', 'bar')
        template_path = os.path.join(template_dir, 'path')
        template_str = self.example_yaml_template()

        resource_type1_class = mock.Mock()
        resource_type1_instance = mock.Mock()
        resource_type1_class.return_value = resource_type1_instance
        resource_type2_class = mock.Mock()
        resource_type2_instance = mock.Mock()
        resource_type2_class.return_value = resource_type2_instance

        resources_to_export = {
            "resource_type1": resource_type1_class,
            "resource_type2": resource_type2_class,
        }
        properties1 = {
            "foo": "bar",
            "Fn::Transform": {
                "Name": "AWS::Include",
                "Parameters": {"Location": "foo.yaml"},
            },
        }
        properties2 = {
            "foo": "bar",
            "Fn::Transform": {"Name": "AWS::OtherTransform"},
        }
        properties_in_list = {
            "Fn::Transform": {
                "Name": "AWS::Include",
                "Parameters": {"Location": "bar.yaml"},
            }
        }
        template_dict = {
            "Resources": {
                "Resource1": {
                    "Type": "resource_type1",
                    "Properties": properties1,
                },
                "Resource2": {
                    "Type": "resource_type2",
                    "Properties": properties2,
                },
            },
            "List": ["foo", properties_in_list],
        }
        open_mock = mock.mock_open()
        include_transform_export_handler_mock = mock.Mock()
        include_transform_export_handler_mock.return_value = {
            "Name": "AWS::Include",
            "Parameters": {"Location": "s3://foo"},
        }
        yaml_parse_mock.return_value = template_dict

        with mock.patch(
            "awscli.customizations.cloudformation."
            "artifact_exporter.compat_open",
            open_mock(read_data=template_str),
        ) as open_mock:
            with mock.patch.dict(
                GLOBAL_EXPORT_DICT,
                {"Fn::Transform": include_transform_export_handler_mock},
            ):
                template_exporter = Template(
                    template_path,
                    parent_dir,
                    self.s3_uploader_mock,
                    resources_to_export,
                )

                exported_template = template_exporter.export_global_artifacts(
                    template_exporter.template_dict
                )

                first_call_args, kwargs = (
                    include_transform_export_handler_mock.call_args_list[0]
                )
                second_call_args, kwargs = (
                    include_transform_export_handler_mock.call_args_list[1]
                )
                third_call_args, kwargs = (
                    include_transform_export_handler_mock.call_args_list[2]
                )
                call_args = [
                    first_call_args[0],
                    second_call_args[0],
                    third_call_args[0],
                ]
                self.assertTrue(
                    {
                        "Name": "AWS::Include",
                        "Parameters": {"Location": "foo.yaml"},
                    }
                    in call_args
                )
                self.assertTrue({"Name": "AWS::OtherTransform"} in call_args)
                self.assertTrue(
                    {
                        "Name": "AWS::Include",
                        "Parameters": {"Location": "bar.yaml"},
                    }
                    in call_args
                )
                self.assertTrue(template_dir in first_call_args)
                self.assertTrue(template_dir in second_call_args)
                self.assertTrue(template_dir in third_call_args)
                self.assertEqual(
                    include_transform_export_handler_mock.call_count, 3
                )
                # new s3 url is added to include location
                self.assertEqual(
                    exported_template["Resources"]["Resource1"]["Properties"][
                        "Fn::Transform"
                    ],
                    {
                        "Name": "AWS::Include",
                        "Parameters": {"Location": "s3://foo"},
                    },
                )
                self.assertEqual(
                    exported_template["List"][1]["Fn::Transform"],
                    {
                        "Name": "AWS::Include",
                        "Parameters": {"Location": "s3://foo"},
                    },
                )

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.is_local_file"
    )
    def test_include_transform_export_handler_with_relative_file_path(
        self, is_local_file_mock
    ):
        # exports transform
        parent_dir = os.path.abspath("someroot")
        self.s3_uploader_mock.upload_with_dedup.return_value = "s3://foo"
        is_local_file_mock.return_value = True
        abs_file_path = os.path.join(parent_dir, "foo.yaml")

        handler_output = include_transform_export_handler(
            {"Name": "AWS::Include", "Parameters": {"Location": "foo.yaml"}},
            self.s3_uploader_mock,
            parent_dir,
        )
        self.s3_uploader_mock.upload_with_dedup.assert_called_once_with(
            abs_file_path
        )
        is_local_file_mock.assert_called_with(abs_file_path)
        self.assertEqual(
            handler_output,
            {'Name': 'AWS::Include', 'Parameters': {'Location': 's3://foo'}},
        )

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.is_local_file"
    )
    def test_include_transform_export_handler_with_absolute_file_path(
        self, is_local_file_mock
    ):
        # exports transform
        parent_dir = os.path.abspath("someroot")
        self.s3_uploader_mock.upload_with_dedup.return_value = "s3://foo"
        is_local_file_mock.return_value = True
        abs_file_path = os.path.abspath(os.path.join("my", "file.yaml"))

        handler_output = include_transform_export_handler(
            {
                "Name": "AWS::Include",
                "Parameters": {"Location": abs_file_path},
            },
            self.s3_uploader_mock,
            parent_dir,
        )
        self.s3_uploader_mock.upload_with_dedup.assert_called_once_with(
            abs_file_path
        )
        is_local_file_mock.assert_called_with(abs_file_path)
        self.assertEqual(
            handler_output,
            {'Name': 'AWS::Include', 'Parameters': {'Location': 's3://foo'}},
        )

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.is_local_file"
    )
    def test_include_transform_export_handler_with_s3_uri(
        self, is_local_file_mock
    ):
        handler_output = include_transform_export_handler(
            {
                "Name": "AWS::Include",
                "Parameters": {"Location": "s3://bucket/foo.yaml"},
            },
            self.s3_uploader_mock,
            "parent_dir",
        )
        # Input is returned unmodified
        self.assertEqual(
            handler_output,
            {
                "Name": "AWS::Include",
                "Parameters": {"Location": "s3://bucket/foo.yaml"},
            },
        )

        is_local_file_mock.assert_not_called()
        self.s3_uploader_mock.assert_not_called()

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.is_local_file"
    )
    def test_include_transform_export_handler_with_no_path(
        self, is_local_file_mock
    ):
        handler_output = include_transform_export_handler(
            {"Name": "AWS::Include", "Parameters": {"Location": ""}},
            self.s3_uploader_mock,
            "parent_dir",
        )
        # Input is returned unmodified
        self.assertEqual(
            handler_output,
            {"Name": "AWS::Include", "Parameters": {"Location": ""}},
        )

        is_local_file_mock.assert_not_called()
        self.s3_uploader_mock.assert_not_called()

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.is_local_file"
    )
    def test_include_transform_export_handler_with_dict_value_for_location(
        self, is_local_file_mock
    ):
        handler_output = include_transform_export_handler(
            {
                "Name": "AWS::Include",
                "Parameters": {
                    "Location": {"Fn::Sub": "${S3Bucket}/file.txt"}
                },
            },
            self.s3_uploader_mock,
            "parent_dir",
        )
        # Input is returned unmodified
        self.assertEqual(
            handler_output,
            {
                "Name": "AWS::Include",
                "Parameters": {
                    "Location": {"Fn::Sub": "${S3Bucket}/file.txt"}
                },
            },
        )

        is_local_file_mock.assert_not_called()
        self.s3_uploader_mock.assert_not_called()

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.is_local_file"
    )
    def test_include_transform_export_handler_non_local_file(
        self, is_local_file_mock
    ):
        # returns unchanged template dict if transform not a local file, and not a S3 URI
        is_local_file_mock.return_value = False

        with self.assertRaises(exceptions.InvalidLocalPathError):
            include_transform_export_handler(
                {
                    "Name": "AWS::Include",
                    "Parameters": {"Location": "http://foo.yaml"},
                },
                self.s3_uploader_mock,
                "parent_dir",
            )
            is_local_file_mock.assert_called_with("http://foo.yaml")
            self.s3_uploader_mock.assert_not_called()

    @mock.patch(
        "awscli.customizations.cloudformation.artifact_exporter.is_local_file"
    )
    def test_include_transform_export_handler_non_include_transform(
        self, is_local_file_mock
    ):
        # ignores transform that is not aws::include
        handler_output = include_transform_export_handler(
            {
                "Name": "AWS::OtherTransform",
                "Parameters": {"Location": "foo.yaml"},
            },
            self.s3_uploader_mock,
            "parent_dir",
        )
        self.s3_uploader_mock.upload_with_dedup.assert_not_called()
        self.assertEqual(
            handler_output,
            {
                "Name": "AWS::OtherTransform",
                "Parameters": {"Location": "foo.yaml"},
            },
        )

    def test_template_export_path_be_folder(self):
        template_path = "/path/foo"
        # Set parent_dir to be a non-existent folder
        with self.assertRaises(ValueError):
            Template(template_path, "somefolder", self.s3_uploader_mock)

        # Set parent_dir to be a real folder, but just a relative path
        with self.make_temp_dir() as dirname:
            with self.assertRaises(ValueError):
                Template(
                    template_path,
                    os.path.relpath(dirname),
                    self.s3_uploader_mock,
                )

    def test_make_zip(self):
        test_file_creator = FileCreator()
        test_file_creator.append_file(
            'index.js',
            'exports handler = (event, context, callback) => {callback(null, event);}',
        )

        dirname = test_file_creator.rootdir

        expected_files = set(['index.js'])

        random_name = ''.join(
            random.choice(string.ascii_letters) for _ in range(10)
        )
        outfile = os.path.join(tempfile.gettempdir(), random_name)

        zipfile_name = None
        try:
            zipfile_name = make_zip(outfile, dirname)

            test_zip_file = zipfile.ZipFile(zipfile_name, "r")
            with closing(test_zip_file) as zf:
                files_in_zip = set()
                for info in zf.infolist():
                    files_in_zip.add(info.filename)

                self.assertEqual(files_in_zip, expected_files)

        finally:
            if zipfile_name:
                os.remove(zipfile_name)
            test_file_creator.remove_all()

    @mock.patch("shutil.copy")
    @mock.patch("tempfile.mkdtemp")
    def test_copy_to_temp_dir(self, mkdtemp_mock, copyfile_mock):
        temp_dir = "/tmp/foo/"
        filename = "test.js"
        mkdtemp_mock.return_value = temp_dir

        returned_dir = copy_to_temp_dir(filename)

        self.assertEqual(returned_dir, temp_dir)
        copyfile_mock.assert_called_once_with(filename, temp_dir + filename)

    @contextmanager
    def make_temp_dir(self):
        filename = tempfile.mkdtemp()
        try:
            yield filename
        finally:
            if filename:
                os.rmdir(filename)

    def example_yaml_template(self):
        return """
        AWSTemplateFormatVersion: '2010-09-09'
        Description: Simple CRUD webservice. State is stored in a SimpleTable (DynamoDB) resource.
        Resources:
        MyFunction:
          Type: AWS::Lambda::Function
          Properties:
            Code: ./handler
            Handler: index.get
            Role:
              Fn::GetAtt:
              - MyFunctionRole
              - Arn
            Timeout: 20
            Runtime: nodejs4.3
        """



# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from urllib.parse import urlencode

import pytest
from botocore.session import Session
from botocore.signers import RequestSigner

from awscli.customizations.ec2instanceconnect.eicesigner import (
    InstanceConnectEndpointRequestSigner,
)
from awscli.testutils import mock


@pytest.fixture
def service_model_mock():
    return mock.Mock()


@pytest.fixture
def session_mock(service_model_mock):
    session_mock = mock.Mock(spec=Session)
    session_mock.get_service_model.return_value = service_model_mock
    return session_mock


@pytest.fixture
def request_signer_mock():
    return mock.Mock(spec=RequestSigner)


@pytest.fixture
def eice_id():
    return "instanceConnectEndpointId"


@pytest.fixture
def remote_port():
    return "1"


@pytest.fixture
def private_ip():
    return "127.0.0.1"


@pytest.fixture
def dns_name():
    return "amazon.com"


@pytest.fixture
def max_tunnel_duration():
    return "1"


class TestInstanceConnectEndpointRequestSigner:
    def test_get_presigned_url_without_max_tunnel_duration(
        self,
        session_mock,
        request_signer_mock,
        eice_id,
        remote_port,
        private_ip,
        dns_name,
    ):
        eice_request_signer = InstanceConnectEndpointRequestSigner(
            session_mock,
            instance_connect_endpoint_dns_name=dns_name,
            instance_connect_endpoint_id=eice_id,
            remote_port=remote_port,
            instance_private_ip=private_ip,
            max_tunnel_duration=None,
            request_singer=request_signer_mock,
        )

        expected_url = "wss://" + dns_name + "/openTunnel?params"
        request_signer_mock.generate_presigned_url.return_value = expected_url

        returned_url = eice_request_signer.get_presigned_url()

        assert returned_url == expected_url

        qs_components = {
            "instanceConnectEndpointId": eice_id,
            "remotePort": remote_port,
            "privateIpAddress": private_ip,
        }
        expected_query_string = urlencode(qs_components)
        expected_request_dict = {
            "url": "wss://"
            + dns_name
            + "/openTunnel?"
            + expected_query_string,
            "body": {},
            "headers": {"host": dns_name},
            "method": "GET",
            "query_string": "",
            "url_path": "/",
            "context": {},
        }
        request_signer_mock.generate_presigned_url.assert_called_with(
            request_dict=expected_request_dict,
            operation_name="openTunnel",
            expires_in=60,
        )

    def test_get_presigned_url_with_max_connection_timeout(
        self,
        session_mock,
        request_signer_mock,
        eice_id,
        remote_port,
        private_ip,
        dns_name,
        max_tunnel_duration,
    ):
        eice_request_signer = InstanceConnectEndpointRequestSigner(
            session_mock,
            instance_connect_endpoint_dns_name=dns_name,
            instance_connect_endpoint_id=eice_id,
            remote_port=remote_port,
            instance_private_ip=private_ip,
            max_tunnel_duration=max_tunnel_duration,
            request_singer=request_signer_mock,
        )

        expected_url = "wss://" + dns_name + "/openTunnel?params"
        request_signer_mock.generate_presigned_url.return_value = expected_url

        returned_url = eice_request_signer.get_presigned_url()

        assert returned_url == expected_url

        qs_components = {
            "instanceConnectEndpointId": eice_id,
            "remotePort": remote_port,
            "privateIpAddress": private_ip,
            "maxTunnelDuration": max_tunnel_duration,
        }
        expected_query_string = urlencode(qs_components)
        expected_request_dict = {
            "url": "wss://"
            + dns_name
            + "/openTunnel?"
            + expected_query_string,
            "body": {},
            "headers": {"host": dns_name},
            "method": "GET",
            "query_string": "",
            "url_path": "/",
            "context": {},
        }
        eice_request_signer._request_signer.generate_presigned_url.assert_called_with(
            request_dict=expected_request_dict,
            operation_name="openTunnel",
            expires_in=60,
        )



#!/usr/bin/env python
# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest


class TestGetHostedZone(BaseAWSCommandParamsTest):
    prefix = 'route53 list-resource-record-sets'

    def test_no_pagination_args(self):
        args = ' --hosted-zone-id /hostedzone/ZD3IYMVP1KDDM'
        cmdline = self.prefix + args
        self.assert_params_for_cmd(
            cmdline, {'HostedZoneId': 'ZD3IYMVP1KDDM'}, expected_rc=0
        )

    def test_with_max_items_pagination(self):
        args = ' --hosted-zone-id /hostedzone/ZD3IYMVP1KDDM --max-items 1'
        cmdline = self.prefix + args
        # We don't map to the service's max-items
        self.assert_params_for_cmd(
            cmdline, {'HostedZoneId': 'ZD3IYMVP1KDDM'}, expected_rc=0
        )

    def test_with_max_override_starting_args(self):
        args = (
            ' --hosted-zone-id /hostedzone/ZD3IYMVP1KDDM'
            ' --max-items 1'
            ' --start-record-name foo'
        )
        cmdline = self.prefix + args
        # Here we _should_ be mapping to the service's arguments
        # because --start-record-name triggered the disabling of
        # pagination.
        self.assert_params_for_cmd(
            cmdline,
            {
                'HostedZoneId': 'ZD3IYMVP1KDDM',
                'StartRecordName': 'foo',
                'MaxItems': '1',
            },
            expected_rc=0,
        )



#!/usr/bin/env python
# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Test help output for the AWS CLI.

The purpose of these docs is to test that the generated output looks how
we expect.

It's intended to be as end to end as possible, but instead of looking
at the man output, we look one step before at the generated rst output
(it's easier to verify).

"""

import os
import re
import shlex

import docutils.nodes
import docutils.parsers.rst
import docutils.utils
import pytest

from awscli.argparser import ArgTableArgParser, MainArgParser, ServiceArgParser
from awscli.testutils import BaseAWSHelpOutputTest, create_clidriver

# Mapping of command names to subcommands that have examples in their help
# output.  This isn't mean to be an exhaustive list, but should help catch
# things like command table renames, virtual commands, etc.
COMMAND_EXAMPLES = {
    'cloudwatch': ['put-metric-data'],
    's3': ['cp', 'ls', 'mb', 'mv', 'rb', 'rm', 'sync'],
    's3api': ['get-object', 'put-object'],
    'ec2': ['run-instances', 'start-instances', 'stop-instances'],
    'swf': ['deprecate-domain', 'describe-domain'],
    'sqs': ['create-queue', 'get-queue-attributes'],
    'emr': [
        'add-steps',
        'create-default-roles',
        'describe-cluster',
        'schedule-hbase-backup',
    ],
}
_dname = os.path.dirname
EXAMPLES_DIR = os.path.join(
    _dname(_dname(_dname(_dname(os.path.abspath(__file__))))),
    'awscli',
    'examples',
)

ALLOWED_FILENAME_CHAR_REGEX = re.compile(r'([a-z0-9_\-\.]*$)')
HTTP_LINK_REGEX = re.compile(r'`.+?<http://')


# Used so that docutils doesn't write errors to stdout/stderr.
# We're collecting and reporting these via AssertionErrors messages.
class NoopWriter:
    def write(self, *args, **kwargs):
        pass


class _ExampleTests(BaseAWSHelpOutputTest):
    def noop_test(self):
        pass


def _get_example_test_cases():
    test_cases = []
    for command, subcommands in COMMAND_EXAMPLES.items():
        for subcommand in subcommands:
            test_cases.append((command, subcommand))
    return test_cases


def _get_all_doc_examples():
    rst_doc_examples = []
    other_doc_examples = []
    # Iterate over all rst doc examples
    for root, _, filenames in os.walk(EXAMPLES_DIR):
        for filename in filenames:
            full_path = os.path.join(root, filename)
            if filename.startswith('.'):
                # Ignore hidden files as it starts with "."
                continue
            if not filename.endswith('.rst'):
                other_doc_examples.append(full_path)
                continue
            rst_doc_examples.append(full_path)
    return rst_doc_examples, other_doc_examples


RST_DOC_EXAMPLES, OTHER_DOC_EXAMPLES = _get_all_doc_examples()
EXAMPLE_COMMAND_TESTS = _get_example_test_cases()


def line_num(content, loc):
    return content[:loc].count('\n') + 1


def extract_error_line(content, error_start, error_end):
    error_line_begin = content.rfind('\n', 0, error_start)
    error_line_end = content.find('\n', error_end)
    return content[error_line_begin:error_line_end]


@pytest.fixture(scope="module")
def command_validator():
    # CLIDriver can take up a lot of resources so we'll just create one
    # instance and use it for all the validation tests.
    driver = create_clidriver()
    return CommandValidator(driver)


@pytest.mark.parametrize("command, subcommand", EXAMPLE_COMMAND_TESTS)
def test_examples(command, subcommand):
    t = _ExampleTests(methodName='noop_test')
    t.setUp()
    try:
        t.driver.main([command, subcommand, 'help'])
        t.assert_contains_with_count('========\nExamples\n========', 1)
    finally:
        t.tearDown()


@pytest.mark.filterwarnings('ignore::DeprecationWarning')
@pytest.mark.parametrize("example_file", RST_DOC_EXAMPLES)
def test_rst_doc_examples(command_validator, example_file):
    verify_has_only_ascii_chars(example_file)
    verify_is_valid_rst(example_file)
    verify_cli_commands_valid(example_file, command_validator)
    verify_no_http_links(example_file)


def verify_no_http_links(filename):
    with open(filename) as f:
        contents = f.read()
    match = HTTP_LINK_REGEX.search(contents)
    if match:
        error_line_number = line_num(contents, match.span()[0])
        error_line = extract_error_line(
            contents, match.span()[0], match.span()[1]
        )
        marker_idx = error_line.find('http://') - 1
        marker_line = (" " * marker_idx) + '^'
        raise AssertionError(
            'Found http:// link in the examples file %s, line %s\n'
            '%s\n%s' % (filename, error_line_number, error_line, marker_line)
        )


def verify_has_only_ascii_chars(filename):
    with open(filename, 'rb') as f:
        bytes_content = f.read()
        try:
            bytes_content.decode('ascii')
        except UnicodeDecodeError as e:
            # The test has failed so we'll try to provide a useful error
            # message.
            offset = e.start
            spread = 20
            bad_text = bytes_content[offset - spread : e.start + spread]
            underlined = ' ' * spread + '^'
            error_text = '\n'.join([bad_text, underlined])
            line_number = bytes_content[:offset].count(b'\n') + 1
            raise AssertionError(
                "Non ascii characters found in the examples file %s, line %s:"
                "\n\n%s\n" % (filename, line_number, error_text)
            )


def verify_is_valid_rst(filename):
    _, errors = parse_rst(filename)
    if errors:
        raise AssertionError(_make_error_msg(filename, errors))


def parse_rst(filename):
    with open(filename) as f:
        contents = f.read()
    parser = docutils.parsers.rst.Parser()
    components = (docutils.parsers.rst.Parser,)
    settings = docutils.frontend.OptionParser(
        components=components
    ).get_default_values()
    document = docutils.utils.new_document('<cli-example>', settings=settings)
    errors = []

    def record_errors(msg):
        msg.level = msg['level']
        msg.type = msg['type']
        error_message = docutils.nodes.Element.astext(msg.children[0])
        line_number = msg['line']
        errors.append({'msg': error_message, 'line_number': line_number})

    document.reporter.stream = NoopWriter()
    document.reporter.attach_observer(record_errors)
    parser.parse(contents, document)
    return document, errors


def _make_error_msg(filename, errors):
    with open(filename) as f:
        lines = f.readlines()
    relative_name = filename[len(EXAMPLES_DIR) + 1 :]
    failure_message = [
        'The file "%s" contains invalid RST: ' % relative_name,
        '',
    ]
    for error in errors:
        # This may not always be super helpful because you sometimes need
        # more than one line of context to understand what went wrong,
        # but by giving you the filename and the line number, it's usually
        # enough to track down what went wrong.
        line_number = min(error['line_number'], len(lines))
        line_number -= 1
        if line_number > 0:
            line_number -= 1
        current_message = [
            'Line %s: %s' % (error['line_number'], error['msg']),
            '  %s' % lines[line_number],
        ]
        failure_message.extend(current_message)
    return '\n'.join(failure_message)


def verify_cli_commands_valid(filename, command_validator):
    cli_commands = find_all_cli_commands(filename)
    for command in cli_commands:
        command_validator.validate_cli_command(command, filename)


def find_all_cli_commands(filename):
    document, _ = parse_rst(filename)
    visitor = CollectCLICommands(document)
    document.walk(visitor)
    return visitor.cli_commands


class CommandValidator:
    def __init__(self, driver):
        self.driver = driver
        help_command = self.driver.create_help_command()
        self._service_command_table = help_command.command_table
        self._global_arg_table = help_command.arg_table
        self._main_parser = MainArgParser(
            self._service_command_table,
            driver.session.user_agent(),
            'Some description',
            self._global_arg_table,
            prog="aws",
        )

    def validate_cli_command(self, command, filename):
        """
        Validates the syntax of a CLI command by actually parsing it, which catches:
        - Invalid service names
        - Invalid operation names
        - Missing required arguments
        - Invalid argument names
        - Malformed argument syntax
        """
        try:
            command_parts = shlex.split(command)[1:]
            # Strip newlines from arguments that may have been included due to
            # backslash line continuations in the RST examples
            command_parts = [part.lstrip('\n') for part in command_parts]
        except Exception as e:
            raise AssertionError(
                "Failed to parse this example as shell command: %s\n\n"
                "Error:\n%s\n" % (command, e)
            )

        # TODO - for now skip validation if command uses
        # --cli-input-json, --cli-input-yaml, or --generate-cli-skeleton
        if any(
            arg.startswith('--cli-input-json')
            or arg.startswith('--cli-input-yaml')
            or arg == '--generate-cli-skeleton'
            for arg in command_parts
        ):
            return

        try:
            # Parse global args
            parsed_args, remaining = self._main_parser.parse_known_args(
                command_parts
            )
            service_name = parsed_args.command

            # Get the service command
            service_cmd = self._service_command_table[service_name]

            # Create the service parser
            cmd_table = service_cmd.create_help_command().command_table
            if not cmd_table:
                # Top-level commands without sub-operations (e.g. login)
                return
            service_parser = ServiceArgParser(
                operations_table=cmd_table, service_name=service_name
            )

            # Parse the operation
            parsed_operation, remaining_args = service_parser.parse_known_args(
                remaining
            )

            if not hasattr(parsed_operation, 'operation'):
                # Not a standard operation (might be help, wait, etc.)
                return

            operation_name = parsed_operation.operation
            operation_cmd = cmd_table[operation_name]

            # Validate all arguments using the operation's argument parser
            arg_table = operation_cmd.arg_table
            operation_parser = ArgTableArgParser(arg_table)

            errors = []
            operation_parser._print_message = (
                lambda message, file: errors.append(message)
            )

            try:
                operation_parser.parse_known_args(remaining_args)
            except (SystemExit, Exception) as e:
                error_msg = [
                    f'Invalid CLI command: {command}\n\n',
                    f'File: {filename}\n\n',
                ]
                if errors:
                    error_msg.extend(errors)
                else:
                    error_msg.append(f'Error: {e}\n')
                raise AssertionError(''.join(error_msg))

        except Exception as e:
            raise AssertionError(
                f"Failed to validate command: {command}\n"
                f"File: {filename}\n"
                f"Error: {e}"
            )


class CollectCLICommands(docutils.nodes.GenericNodeVisitor):
    def __init__(self, document):
        docutils.nodes.GenericNodeVisitor.__init__(self, document)
        self.cli_commands = []

    def visit_literal_block(self, node):
        contents = node.rawsource.strip()
        if contents.startswith('aws '):
            self.cli_commands.append(contents)

    def default_visit(self, node):
        pass


@pytest.mark.parametrize("example_file", RST_DOC_EXAMPLES + OTHER_DOC_EXAMPLES)
def test_example_file_name(example_file):
    filename = example_file.split(os.sep)[-1]
    _assert_file_is_rst_or_txt(example_file)
    _assert_name_contains_only_allowed_characters(filename)


def _assert_file_is_rst_or_txt(filepath):
    assert filepath.endswith('.rst') or filepath.endswith('.txt')


def _assert_name_contains_only_allowed_characters(filename):
    assert ALLOWED_FILENAME_CHAR_REGEX.match(filename)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.commands import BasicCommand, BasicHelp
from awscli.customizations.exceptions import ParamValidationError
from awscli.customizations.wizard.loader import WizardLoader


def register_wizard_commands(event_handlers):
    from awscli.customizations.wizard import devcommands

    devcommands.register_dev_commands(event_handlers)
    loader = WizardLoader()
    commands = loader.list_commands_with_wizards()
    _register_wizards_for_commands(commands, event_handlers)


def _register_wizards_for_commands(commands, event_handlers):
    for command in commands:
        event_handlers.register(
            f'building-command-table.{command}', _add_wizard_command
        )


def _add_wizard_command(session, command_object, command_table, **kwargs):
    cmd = TopLevelWizardCommand(
        session=session,
        loader=WizardLoader(),
        parent_command=command_object.name,
    )
    command_table['wizard'] = cmd


class TopLevelWizardCommand(BasicCommand):
    NAME = 'wizard'
    DESCRIPTION = (
        'Interactive command for creating and configuring AWS resources.'
    )

    def __init__(
        self, session, loader, parent_command, runner=None, wizard_name='_main'
    ):
        super().__init__(session)
        self._session = session
        self._loader = loader
        self._parent_command = parent_command
        self._runner = runner
        self._wizard_name = wizard_name

    def _get_runner(self):
        # If a runner was not provided during initialization, compute the
        # default. This defers default computation to runtime, when the
        # wizard command is actually invoked. The benefit of this is so that
        # we defer importing `awscli.customizations.wizard.factory` until
        # it's actually needed. The wizard factory module imports from
        # `prompt_toolkit`, and importing `prompt_toolkit` while executing
        # commands that don't actually use it has historically led to
        # unnecessarily higher command execution time and wasted compute.
        if self._runner is None:
            from awscli.customizations.wizard import factory

            self._runner = {
                '0.1': factory.create_default_wizard_v1_runner(self._session),
                '0.2': factory.create_default_wizard_v2_runner(self._session),
            }
        return self._runner

    def _build_subcommand_table(self):
        subcommand_table = super()._build_subcommand_table()
        wizards = self._get_available_wizards()
        for name in wizards:
            cmd = SingleWizardCommand(
                self._session,
                self._loader,
                self._parent_command,
                runner=self._runner,
                wizard_name=name,
            )
            subcommand_table[name] = cmd
        self._add_lineage(subcommand_table)
        return subcommand_table

    def _get_available_wizards(self):
        wizards = self._loader.list_available_wizards(self._parent_command)
        return [name for name in wizards if not name.startswith('_')]

    def _run_main(self, parsed_args, parsed_globals):
        if self._wizard_exists():
            self._run_wizard()
            return 0
        else:
            self._raise_usage_error()

    def _wizard_exists(self):
        return self._loader.wizard_exists(
            self._parent_command, self._wizard_name
        )

    def _run_wizard(self):
        loaded = self._loader.load_wizard(
            self._parent_command, self._wizard_name
        )
        version = loaded.get('version')
        runner = self._get_runner()
        if version in runner:
            runner[version].run(loaded)
        else:
            raise ParamValidationError(
                f'Definition file has unsupported version {version} '
            )

    def create_help_command(self):
        return BasicHelp(
            self._session,
            self,
            command_table=self.subcommand_table,
            arg_table=self.arg_table,
        )


class SingleWizardCommand(TopLevelWizardCommand):
    def __init__(self, session, loader, parent_command, runner, wizard_name):
        super().__init__(
            session,
            loader,
            parent_command,
            runner=runner,
            wizard_name=wizard_name,
        )
        self._session = session
        self._loader = loader
        self._runner = runner
        self._wizard_name = wizard_name
        self.NAME = self._wizard_name

    def _build_subcommand_table(self):
        return {}

    def _run_main(self, parsed_args, parsed_globals):
        self._run_wizard()
        return 0

    def create_help_command(self):
        loaded = self._loader.load_wizard(
            self._parent_command,
            self._wizard_name,
        )
        return WizardHelpCommand(
            self._session, self, self.subcommand_table, self.arg_table, loaded
        )


class WizardHelpCommand(BasicHelp):
    def __init__(
        self, session, command_object, command_table, arg_table, loaded_wizard
    ):
        super().__init__(session, command_object, command_table, arg_table)
        self._description = loaded_wizard.get('description', '')



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest

CHANGEBATCH_JSON = (
    '{"Comment":"string","Changes":['
    '{"Action":"CREATE","ResourceRecordSet":{'
    '"Name":"test-foo.bar.com",'
    '"Type":"CNAME",'
    '"TTL":300,'
    '"ResourceRecords":['
    '{"Value":"foo-bar-com"}'
    ']}}]}'
)


class TestGetHostedZone(BaseAWSCommandParamsTest):
    prefix = 'route53 get-hosted-zone'

    def setUp(self):
        super(TestGetHostedZone, self).setUp()

    def test_full_resource_id(self):
        args = ' --id /hostedzone/ZD3IYMVP1KDDM'
        cmdline = self.prefix + args
        self.assert_params_for_cmd(
            cmdline, {'Id': 'ZD3IYMVP1KDDM'}, expected_rc=0
        )

    def test_short_resource_id(self):
        args = ' --id ZD3IYMVP1KDDM'
        cmdline = self.prefix + args
        self.assert_params_for_cmd(
            cmdline, {'Id': 'ZD3IYMVP1KDDM'}, expected_rc=0
        )


class TestChangeResourceRecord(BaseAWSCommandParamsTest):
    prefix = 'route53 change-resource-record-sets'

    def setUp(self):
        super(TestChangeResourceRecord, self).setUp()

    def test_full_resource_id(self):
        args = ' --hosted-zone-id /change/ZD3IYMVP1KDDM'
        args += ' --change-batch %s' % CHANGEBATCH_JSON
        cmdline = self.prefix + args
        expected = {
            "HostedZoneId": "ZD3IYMVP1KDDM",
            "ChangeBatch": {
                "Comment": "string",
                "Changes": [
                    {
                        "Action": "CREATE",
                        "ResourceRecordSet": {
                            "Name": "test-foo.bar.com",
                            "Type": "CNAME",
                            "TTL": 300,
                            "ResourceRecords": [{"Value": "foo-bar-com"}],
                        },
                    }
                ],
            },
        }
        self.assert_params_for_cmd(cmdline, expected, expected_rc=0)


class TestGetChange(BaseAWSCommandParamsTest):
    prefix = 'route53 get-change'

    def setUp(self):
        super(TestGetChange, self).setUp()

    def test_full_resource_id(self):
        args = ' --id /change/ZD3IYMVP1KDDM'
        cmdline = self.prefix + args
        expected = {'Id': 'ZD3IYMVP1KDDM'}
        self.assert_params_for_cmd(cmdline, expected, expected_rc=0)

    def test_short_resource_id(self):
        args = ' --id ZD3IYMVP1KDDM'
        cmdline = self.prefix + args
        expected = {'Id': 'ZD3IYMVP1KDDM'}
        self.assert_params_for_cmd(cmdline, expected, expected_rc=0)


class TestReusableDelegationSet(BaseAWSCommandParamsTest):
    prefix = 'route53 get-reusable-delegation-set'

    def setUp(self):
        super(TestReusableDelegationSet, self).setUp()

    def test_full_resource_id(self):
        args = ' --id /delegationset/N9INWVYQ6Q0FN'
        cmdline = self.prefix + args
        self.assert_params_for_cmd(
            cmdline, {'Id': 'N9INWVYQ6Q0FN'}, expected_rc=0
        )

    def test_short_resource_id(self):
        args = ' --id N9INWVYQ6Q0FN'
        cmdline = self.prefix + args
        self.assert_params_for_cmd(
            cmdline, {'Id': 'N9INWVYQ6Q0FN'}, expected_rc=0
        )


class TestMaxItems(BaseAWSCommandParamsTest):
    prefix = 'route53 list-resource-record-sets'

    def test_full_resource_id(self):
        args = ' --hosted-zone-id /hostedzone/ABCD --max-items 1'
        cmdline = self.prefix + args
        expected = {'HostedZoneId': 'ABCD'}
        self.assert_params_for_cmd(cmdline, expected, expected_rc=0)



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseCLIDriverTest, create_bucket, mock


class TestCreateBucket(BaseCLIDriverTest):
    def test_bucket_already_owned_by_you(self):
        # TODO: fix this patch when we have a better way to stub out responses
        with mock.patch('botocore.endpoint.Endpoint._send') as _send:
            _send.side_effect = [
                mock.Mock(status_code=500, headers={}, content=b''),
                mock.Mock(
                    status_code=409,
                    headers={},
                    content=b'''<?xml version="1.0" encoding="UTF-8"?>
                        <Error>
                        <Code>BucketAlreadyOwnedByYou</Code>
                        <Message>Your previous request to create the named
                            bucket succeeded and you already own it.</Message>
                        <BucketName>awscli-foo-bar</BucketName>
                        <RequestId>0123456789ABCDEF</RequestId>
                        <HostId>foo</HostId>
                        </Error>''',
                ),
                mock.Mock(status_code=200, headers={}, content=b''),
                mock.Mock(status_code=200, headers={}, content=b''),
            ]
            self.assertEqual(create_bucket(self.session, 'bucket'), 'bucket')



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.s3.syncstrategy.register import (
    register_sync_strategy,
)
from awscli.testutils import mock, unittest


class TestRegisterSyncStrategy(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock()
        self.strategy_cls = mock.Mock()
        self.strategy_object = mock.Mock()
        self.strategy_cls.return_value = self.strategy_object

    def test_register_sync_strategy(self):
        """
        Ensure that registering a single strategy class works as expected
        when ``sync_type`` is specified.
        """
        register_sync_strategy(self.session, self.strategy_cls, 'sync_type')
        # Ensure sync strategy class is instantiated
        self.strategy_cls.assert_called_with('sync_type')
        # Ensure the sync strategy's ``register_strategy`` method is
        # called correctly.
        self.strategy_object.register_strategy.assert_called_with(self.session)

    def test_register_sync_strategy_default_sync_type(self):
        """
        Ensure that registering a single strategy class works as expected
        when the ``sync_type`` is not specified.
        """
        register_sync_strategy(self.session, self.strategy_cls)
        # Ensure sync strategy class is instantiated
        self.strategy_cls.assert_called_with('file_at_src_and_dest')
        # Ensure the sync strategy's ``register_strategy`` method is
        # called correctly.
        self.strategy_object.register_strategy.assert_called_with(self.session)


if __name__ == "__main__":
    unittest.main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.customizations.ecs.exceptions import MissingPropertyError
from awscli.customizations.ecs.filehelpers import (
    APP_PREFIX,
    DGP_PREFIX,
    MAX_CHAR_LENGTH,
    find_required_key,
    get_app_name,
    get_cluster_name_from_arn,
    get_deploy_group_name,
    parse_appspec,
)
from awscli.testutils import unittest


class TestFilehelpers(unittest.TestCase):
    YAML_APPSPEC = """
    version: 0.0
    resources:
      - TestService:
          type: AWS::ECS::Service
          properties:
            taskDefinition: arn:aws:ecs:::task-definition/test:1
            loadBalancerInfo:
              containerName: web
              containerPort: 80
    """

    PARSED_APPSPEC = {
        "version": 0.0,
        "resources": [
            {
                "TestService": {
                    "type": "AWS::ECS::Service",
                    "properties": {
                        "taskDefinition": "arn:aws:ecs:::task-definition/test:1",
                        "loadBalancerInfo": {
                            "containerName": "web",
                            "containerPort": 80,
                        },
                    },
                }
            }
        ],
    }

    MIXED_CASE_APPSPEC = {
        "version": 0.0,
        "Resources": [
            {
                "TestService": {
                    "TYPE": "AWS::ECS::Service",
                    "PROperties": {
                        "TaskDefinition": "arn:aws:ecs:::task-definition/test:1",
                        "loadbalancerInfo": {
                            "containerName": "web",
                            "containerPort": 80,
                        },
                    },
                }
            }
        ],
    }

    def test_find_required_key(self):
        test_properties_dict = {
            "TaskDefinition": "arn:aws:ecs:::task-definition/test:1",
            "loadbalancerInfo": {"containerName": "web", "containerPort": 80},
        }
        test_key = 'taskDefinition'
        expected_result = 'TaskDefinition'

        result = find_required_key(
            'task definition', test_properties_dict, test_key
        )
        self.assertEqual(result, expected_result)

    def test_find_required_key_error_missing_key(self):
        invalid_properties_dict = {'name': 'some-lambda-function'}
        test_key = 'taskDefinition'

        with self.assertRaises(MissingPropertyError):
            find_required_key(
                'task definition', invalid_properties_dict, test_key
            )

    def test_find_required_key_error_empty_object(self):
        test_key = 'taskDefinition'

        with self.assertRaises(MissingPropertyError):
            find_required_key('task definition', dict(), test_key)

    def test_get_app_name_long(self):
        cluster = 'ClusterClusterClusterClusterClusterClusterClusterCluster'
        service = 'ServiceServiceServiceServiceServiceServiceServiceService'

        expected = (
            APP_PREFIX
            + cluster[:MAX_CHAR_LENGTH]
            + '-'
            + service[:MAX_CHAR_LENGTH]
        )
        response = get_app_name(service, cluster, None)

        self.assertEqual(expected, response)

    def test_get_app_name_no_cluster(self):
        service = 'test-service'
        expected_name = 'AppECS-default-test-service'

        response = get_app_name(service, None, None)
        self.assertEqual(expected_name, response)

    def test_get_app_name_provided(self):
        cluster = 'test-cluster'
        service = 'test-service'
        app_name = 'test-app'

        response = get_app_name(service, cluster, app_name)
        self.assertEqual(app_name, response)

    def test_get_cluster_name_from_arn(self):
        name = "default"
        arn = 'arn:aws:ecs:::cluster/default'

        response = get_cluster_name_from_arn(arn)
        self.assertEqual(name, response)

    def test_get_deploy_group_name_long(self):
        cluster = 'ClusterClusterClusterClusterClusterClusterClusterCluster'
        service = 'ServiceServiceServiceServiceServiceServiceServiceService'

        expected = (
            DGP_PREFIX
            + cluster[:MAX_CHAR_LENGTH]
            + '-'
            + service[:MAX_CHAR_LENGTH]
        )
        response = get_deploy_group_name(service, cluster, None)

        self.assertEqual(expected, response)

    def test_get_deploy_group_name_no_cluster(self):
        service = 'test-service'
        expected_name = 'DgpECS-default-test-service'

        response = get_deploy_group_name(service, None, None)
        self.assertEqual(expected_name, response)

    def test_get_deploy_group_name_provided(self):
        cluster = 'test-cluster'
        service = 'test-service'
        dgp_name = 'test-deployment-group'

        response = get_deploy_group_name(service, cluster, dgp_name)
        self.assertEqual(dgp_name, response)

    """ def test_get_ecs_suffix_no_cluster(self):
        service = 'myService'
        expected_suffix = 'default-myService'

        output = get_ecs_suffix(service, None)
        self.assertEqual(output, expected_suffix)

    def test_get_ecs_suffix_cluster_provided(self):
        service = 'myService'
        cluster = 'myCluster'
        expected_suffix = 'myCluster-myService'

        output = get_ecs_suffix(service, cluster)
        self.assertEqual(output, expected_suffix) """

    def test_parse_appsec(self):
        output = parse_appspec(self.YAML_APPSPEC)
        self.assertEqual(output, self.PARSED_APPSPEC)

    def test_parse_appsec_json(self):
        output = parse_appspec(str(self.PARSED_APPSPEC))
        self.assertEqual(output, self.PARSED_APPSPEC)



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations import cloudtrail
from awscli.testutils import mock, unittest


class TestCloudTrailPlumbing(unittest.TestCase):
    def test_initialization_registers_injector(self):
        cli = mock.Mock()
        cloudtrail.initialize(cli)
        cli.register.assert_called_with(
            'building-command-table.cloudtrail', cloudtrail.inject_commands
        )

    def test_injection_adds_two_commands_to_cmd_table(self):
        command_table = {}
        session = mock.Mock()
        cloudtrail.inject_commands(command_table, session)
        self.assertIn('validate-logs', command_table)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.commands import BasicCommand


def awscli_initialize(event_emitter):
    event_emitter.register('building-command-table.main', add_plugin_cmd)


def add_plugin_cmd(command_table, session, **kwargs):
    command_table['plugin-test-cmd'] = PluginTestCommand(session)


class PluginTestCommand(BasicCommand):
    def _run_main(self, parsed_args, parsed_globals):
        return 0



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0e
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import hashlib
import os
import random
import shutil
import string
import tempfile

import botocore
import botocore.exceptions
import botocore.session
from botocore.stub import Stubber
from s3transfer import S3Transfer

from awscli.compat import OrderedDict
from awscli.customizations.s3uploader import NoSuchBucketError, S3Uploader
from awscli.testutils import mock, unittest


class TestS3Uploader(unittest.TestCase):
    def setUp(self):
        self._construct_uploader("us-east-1")

    def _construct_uploader(self, region):
        self.s3client = botocore.session.get_session().create_client(
            's3', region_name=region
        )
        self.s3client_stub = Stubber(self.s3client)
        self.transfer_manager_mock = mock.Mock(spec=S3Transfer)
        self.transfer_manager_mock.upload = mock.Mock()
        self.bucket_name = "bucketname"
        self.prefix = None

        self.s3uploader = S3Uploader(
            self.s3client,
            self.bucket_name,
            self.prefix,
            None,
            False,
            self.transfer_manager_mock,
        )

    @mock.patch('os.path.getsize', return_value=1)
    @mock.patch("awscli.customizations.s3uploader.ProgressPercentage")
    def test_upload_successful(self, progress_percentage_mock, get_size_patch):
        file_name = "filename"
        remote_path = "remotepath"
        prefix = "SomePrefix"
        remote_path_with_prefix = f"{prefix}/{remote_path}"
        s3uploader = S3Uploader(
            self.s3client,
            self.bucket_name,
            prefix,
            None,
            False,
            self.transfer_manager_mock,
        )
        expected_upload_url = f"s3://{self.bucket_name}/{prefix}/{remote_path}"

        # Setup mock to fake that file does not exist
        s3uploader.file_exists = mock.Mock()
        s3uploader.file_exists.return_value = False
        # set the metadata used by the uploader when uploading
        artifact_metadata = {"key": "val"}
        s3uploader.artifact_metadata = artifact_metadata

        upload_url = s3uploader.upload(file_name, remote_path)
        self.assertEqual(expected_upload_url, upload_url)

        expected_extra_args = {
            # expected encryption args
            "ServerSideEncryption": "AES256",
            # expected metadata
            "Metadata": artifact_metadata,
        }
        self.transfer_manager_mock.upload.assert_called_once_with(
            file_name,
            self.bucket_name,
            remote_path_with_prefix,
            expected_extra_args,
            mock.ANY,
        )
        s3uploader.file_exists.assert_called_once_with(remote_path_with_prefix)

    @mock.patch('os.path.getsize', return_value=1)
    @mock.patch("awscli.customizations.s3uploader.ProgressPercentage")
    def test_upload_successful_odict(
        self, progress_percentage_mock, get_size_patch
    ):
        file_name = "filename"
        remote_path = "remotepath"
        prefix = "SomePrefix"
        remote_path_with_prefix = f"{prefix}/{remote_path}"
        s3uploader = S3Uploader(
            self.s3client,
            self.bucket_name,
            prefix,
            None,
            False,
            self.transfer_manager_mock,
        )
        expected_upload_url = f"s3://{self.bucket_name}/{prefix}/{remote_path}"

        # Setup mock to fake that file does not exist
        s3uploader.file_exists = mock.Mock()
        s3uploader.file_exists.return_value = False
        # set the metadata used by the uploader when uploading
        artifact_metadata = OrderedDict({"key": "val"})
        s3uploader.artifact_metadata = artifact_metadata

        upload_url = s3uploader.upload(file_name, remote_path)
        self.assertEqual(expected_upload_url, upload_url)

        expected_extra_args = {
            # expected encryption args
            "ServerSideEncryption": "AES256",
            # expected metadata
            "Metadata": artifact_metadata,
        }
        self.transfer_manager_mock.upload.assert_called_once_with(
            file_name,
            self.bucket_name,
            remote_path_with_prefix,
            expected_extra_args,
            mock.ANY,
        )
        s3uploader.file_exists.assert_called_once_with(remote_path_with_prefix)

    @mock.patch("awscli.customizations.s3uploader.ProgressPercentage")
    def test_upload_idempotency(self, progress_percentage_mock):
        file_name = "filename"
        remote_path = "remotepath"

        # Setup mock to fake that file was already uploaded
        self.s3uploader.file_exists = mock.Mock()
        self.s3uploader.file_exists.return_value = True

        self.s3uploader.upload(file_name, remote_path)

        self.transfer_manager_mock.upload.assert_not_called()
        self.s3uploader.file_exists.assert_called_once_with(remote_path)

    @mock.patch('os.path.getsize', return_value=1)
    @mock.patch("awscli.customizations.s3uploader.ProgressPercentage")
    def test_upload_force_upload(
        self, progress_percentage_mock, get_size_patch
    ):
        file_name = "filename"
        remote_path = "remotepath"
        expected_upload_url = f"s3://{self.bucket_name}/{remote_path}"

        # Set ForceUpload = True
        self.s3uploader = S3Uploader(
            self.s3client,
            self.bucket_name,
            self.prefix,
            None,
            True,
            self.transfer_manager_mock,
        )

        # Pretend file already exists
        self.s3uploader.file_exists = mock.Mock()
        self.s3uploader.file_exists.return_value = True

        # Because we forced an update, this should reupload even if file exists
        upload_url = self.s3uploader.upload(file_name, remote_path)
        self.assertEqual(expected_upload_url, upload_url)

        expected_encryption_args = {"ServerSideEncryption": "AES256"}
        self.transfer_manager_mock.upload.assert_called_once_with(
            file_name,
            self.bucket_name,
            remote_path,
            expected_encryption_args,
            mock.ANY,
        )

        # Since ForceUpload=True, we should NEVER do the file-exists check
        self.s3uploader.file_exists.assert_not_called()

    @mock.patch('os.path.getsize', return_value=1)
    @mock.patch("awscli.customizations.s3uploader.ProgressPercentage")
    def test_upload_successful_custom_kms_key(
        self, progress_percentage_mock, get_size_patch
    ):
        file_name = "filename"
        remote_path = "remotepath"
        kms_key_id = "kms_id"
        expected_upload_url = f"s3://{self.bucket_name}/{remote_path}"
        # Set KMS Key Id
        self.s3uploader = S3Uploader(
            self.s3client,
            self.bucket_name,
            self.prefix,
            kms_key_id,
            False,
            self.transfer_manager_mock,
        )

        # Setup mock to fake that file does not exist
        self.s3uploader.file_exists = mock.Mock()
        self.s3uploader.file_exists.return_value = False

        upload_url = self.s3uploader.upload(file_name, remote_path)
        self.assertEqual(expected_upload_url, upload_url)

        expected_encryption_args = {
            "ServerSideEncryption": "aws:kms",
            "SSEKMSKeyId": kms_key_id,
        }
        self.transfer_manager_mock.upload.assert_called_once_with(
            file_name,
            self.bucket_name,
            remote_path,
            expected_encryption_args,
            mock.ANY,
        )
        self.s3uploader.file_exists.assert_called_once_with(remote_path)

    @mock.patch('os.path.getsize', return_value=1)
    @mock.patch("awscli.customizations.s3uploader.ProgressPercentage")
    def test_upload_successful_nobucket(
        self, progress_percentage_mock, get_size_patch
    ):
        file_name = "filename"
        remote_path = "remotepath"

        # Setup mock to fake that file does not exist
        self.s3uploader.file_exists = mock.Mock()
        self.s3uploader.file_exists.return_value = False

        # Setup uploader to return a NOSuchBucket exception
        exception = botocore.exceptions.ClientError(
            {"Error": {"Code": "NoSuchBucket"}}, "OpName"
        )
        self.transfer_manager_mock.upload.side_effect = exception

        with self.assertRaises(NoSuchBucketError):
            self.s3uploader.upload(file_name, remote_path)

    @mock.patch('os.path.getsize', return_value=1)
    @mock.patch("awscli.customizations.s3uploader.ProgressPercentage")
    def test_upload_successful_exceptions(
        self, progress_percentage_mock, get_size_patch
    ):
        file_name = "filename"
        remote_path = "remotepath"

        # Setup mock to fake that file does not exist
        self.s3uploader.file_exists = mock.Mock()
        self.s3uploader.file_exists.return_value = False

        # Raise an unrecognized botocore error
        exception = botocore.exceptions.ClientError(
            {"Error": {"Code": "SomeError"}}, "OpName"
        )
        self.transfer_manager_mock.upload.side_effect = exception

        with self.assertRaises(botocore.exceptions.ClientError):
            self.s3uploader.upload(file_name, remote_path)

        # Some other exception
        self.transfer_manager_mock.upload.side_effect = FloatingPointError()
        with self.assertRaises(FloatingPointError):
            self.s3uploader.upload(file_name, remote_path)

    def test_upload_with_dedup(self):
        checksum = "some md5 checksum"
        filename = "filename"
        extension = "extn"

        self.s3uploader.file_checksum = mock.Mock()
        self.s3uploader.file_checksum.return_value = checksum

        self.s3uploader.upload = mock.Mock()

        self.s3uploader.upload_with_dedup(filename, extension)

        remotepath = f"{checksum}.{extension}"
        self.s3uploader.upload.assert_called_once_with(filename, remotepath)

    def test_file_exists(self):
        key = "some/path"
        expected_params = {"Bucket": self.bucket_name, "Key": key}
        response = {
            "AcceptRanges": "bytes",
            "ContentType": "text/html",
            "LastModified": "Thu, 16 Apr 2015 18:19:14 GMT",
            "ContentLength": 77,
            "VersionId": "null",
            "ETag": "\"30a6ec7e1a9ad79c203d05a589c8b400\"",
            "Metadata": {},
        }

        # Let's pretend file exists
        self.s3client_stub.add_response(
            "head_object", response, expected_params
        )

        with self.s3client_stub:
            self.assertTrue(self.s3uploader.file_exists(key))

        # Let's pretend file does not exist
        self.s3client_stub.add_client_error(
            'head_object', "ClientError", "some error"
        )
        with self.s3client_stub:
            self.assertFalse(self.s3uploader.file_exists(key))

        # Let's pretend some other unknown exception happened
        s3mock = mock.Mock()
        uploader = S3Uploader(s3mock, self.bucket_name)
        s3mock.head_object = mock.Mock()
        s3mock.head_object.side_effect = RuntimeError()

        with self.assertRaises(RuntimeError):
            uploader.file_exists(key)

    def test_file_checksum(self):
        num_chars = 4096 * 5
        data = ''.join(
            random.choice(string.ascii_uppercase) for _ in range(num_chars)
        ).encode('utf-8')
        md5 = hashlib.md5()
        md5.update(data)
        expected_checksum = md5.hexdigest()

        tempdir = tempfile.mkdtemp()
        try:
            filename = os.path.join(tempdir, 'tempfile')
            with open(filename, 'wb') as f:
                f.write(data)

            actual_checksum = self.s3uploader.file_checksum(filename)
            self.assertEqual(expected_checksum, actual_checksum)
        finally:
            shutil.rmtree(tempdir)

    @mock.patch("awscli.customizations.s3uploader.get_md5")
    def test_file_checksum_fips_fallback(self, get_md5_mock):
        num_chars = 4096*5
        data = ''.join(random.choice(string.ascii_uppercase)
                       for _ in range(num_chars)).encode('utf-8')
        checksum = hashlib.sha256(usedforsecurity=False)
        checksum.update(data)
        expected_checksum = checksum.hexdigest()

        tempdir = tempfile.mkdtemp()
        get_md5_mock.side_effect = botocore.exceptions.MD5UnavailableError()
        try:
            filename = os.path.join(tempdir, 'tempfile')
            with open(filename, 'wb') as f:
                f.write(data)

            actual_checksum = self.s3uploader.file_checksum(filename)
            self.assertEqual(expected_checksum, actual_checksum)
        finally:
            shutil.rmtree(tempdir)

    def test_make_url(self):
        path = "Hello/how/are/you"
        expected = f"s3://{self.bucket_name}/{path}"
        self.assertEqual(expected, self.s3uploader.make_url(path))

    def test_to_path_style_s3_url_us_east_1(self):
        key = "path/to/file"
        version = "someversion"
        region = "us-east-1"
        self._construct_uploader(region)

        s3uploader = S3Uploader(self.s3client, self.bucket_name)
        result = s3uploader.to_path_style_s3_url(key, version)
        versioned_url = (
            "https://s3.us-east-1.amazonaws.com/{0}/{1}?versionId={2}"
        )
        self.assertEqual(
            result,
            versioned_url.format(self.bucket_name, key, version),
        )

        # Without versionId, that query parameter should be omitted
        s3uploader = S3Uploader(self.s3client, self.bucket_name, region)
        result = s3uploader.to_path_style_s3_url(key)
        expected_url = "https://s3.us-east-1.amazonaws.com/{0}/{1}"
        self.assertEqual(
            result, expected_url.format(self.bucket_name, key, version)
        )

    def test_to_path_style_s3_url_other_regions(self):
        key = "path/to/file"
        version = "someversion"
        region = "us-west-2"
        self._construct_uploader(region)

        s3uploader = S3Uploader(self.s3client, self.bucket_name, region)
        result = s3uploader.to_path_style_s3_url(key, version)
        self.assertEqual(
            result,
            f"https://s3.{region}.amazonaws.com/{self.bucket_name}/{key}?versionId={version}",
        )

        # Without versionId, that query parameter should be omitted
        s3uploader = S3Uploader(self.s3client, self.bucket_name, region)
        result = s3uploader.to_path_style_s3_url(key)
        self.assertEqual(
            result,
            f"https://s3.{region}.amazonaws.com/{self.bucket_name}/{key}",
        )

    def test_to_path_style_s3_url_china_regions(self):
        key = "path/to/file"
        version = "someversion"
        region = "cn-northwest-1"
        self._construct_uploader(region)

        s3uploader = S3Uploader(self.s3client, self.bucket_name, region)
        result = s3uploader.to_path_style_s3_url(key, version)
        self.assertEqual(
            result,
            f"https://s3.{region}.amazonaws.com.cn/{self.bucket_name}/{key}?versionId={version}",
        )

        # Without versionId, that query parameter should be omitted
        s3uploader = S3Uploader(self.s3client, self.bucket_name, region)
        result = s3uploader.to_path_style_s3_url(key)
        self.assertEqual(
            result,
            f"https://s3.{region}.amazonaws.com.cn/{self.bucket_name}/{key}",
        )

    def test_artifact_metadata_invalid_type(self):
        prefix = "SomePrefix"
        s3uploader = S3Uploader(
            self.s3client,
            self.bucket_name,
            prefix,
            None,
            False,
            self.transfer_manager_mock,
        )
        invalid_metadata = ["key", "val"]
        with self.assertRaises(TypeError):
            s3uploader.artifact_metadata = invalid_metadata



# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import os
import shutil
import tempfile
import time

from botocore.exceptions import ClientError
from botocore.session import Session

from awscli.testutils import aws, random_chars, unittest

S3_READ_POLICY_ARN = 'arn:aws:iam::aws:policy/AmazonS3ReadOnlyAccess'


class TestAssumeRoleCredentials(unittest.TestCase):
    def setUp(self):
        super(TestAssumeRoleCredentials, self).setUp()
        self.environ = os.environ.copy()
        self.parent_session = Session()
        self.iam = self.parent_session.create_client('iam')
        self.sts = self.parent_session.create_client('sts')
        self.tempdir = tempfile.mkdtemp()
        self.config_file = os.path.join(self.tempdir, 'config')

        # A role trust policy that allows the current account to call assume
        # role on itself.
        account_id = self.sts.get_caller_identity()['Account']
        self.role_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {"AWS": "arn:aws:iam::%s:root" % account_id},
                    "Action": "sts:AssumeRole",
                }
            ],
        }

    def tearDown(self):
        super(TestAssumeRoleCredentials, self).tearDown()
        shutil.rmtree(self.tempdir)

    def random_name(self):
        return 'clitest-' + random_chars(10)

    def create_role(self, policy_document, policy_arn=None):
        name = self.random_name()
        response = self.iam.create_role(
            RoleName=name, AssumeRolePolicyDocument=json.dumps(policy_document)
        )
        self.addCleanup(self.iam.delete_role, RoleName=name)
        if policy_arn:
            self.iam.attach_role_policy(RoleName=name, PolicyArn=policy_arn)
            self.addCleanup(
                self.iam.detach_role_policy,
                RoleName=name,
                PolicyArn=policy_arn,
            )
        return response['Role']

    def create_user(self, policy_arns):
        name = self.random_name()
        user = self.iam.create_user(UserName=name)['User']
        self.addCleanup(self.iam.delete_user, UserName=name)

        for arn in policy_arns:
            self.iam.attach_user_policy(UserName=name, PolicyArn=arn)
            self.addCleanup(
                self.iam.detach_user_policy, UserName=name, PolicyArn=arn
            )

        return user

    def create_creds(self, user_name):
        creds = self.iam.create_access_key(UserName=user_name)['AccessKey']
        self.addCleanup(
            self.iam.delete_access_key,
            UserName=user_name,
            AccessKeyId=creds['AccessKeyId'],
        )
        return creds

    def wait_for_assume_role(
        self,
        role_arn,
        access_key,
        secret_key,
        token=None,
        attempts=30,
        delay=10,
        num_success_needed=3,
    ):
        # "Why not use the policy simulator?" you might ask. The answer is
        # that the policy simulator will return success far before you can
        # actually make the calls.
        client = self.parent_session.create_client(
            'sts',
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
            aws_session_token=token,
        )
        attempts_remaining = attempts
        role_session_name = random_chars(10)
        num_success = 0
        while attempts_remaining > 0:
            try:
                result = client.assume_role(
                    RoleArn=role_arn, RoleSessionName=role_session_name
                )
                num_success += 1
                if num_success == num_success_needed:
                    return result['Credentials']
            except ClientError as e:
                code = e.response.get('Error', {}).get('Code')
                if code not in ["InvalidClientTokenId", "AccessDenied"]:
                    raise
                attempts_remaining -= 1
            time.sleep(delay)

        raise Exception("Unable to assume role %s" % role_arn)

    def create_assume_policy(self, role_arn):
        policy_document = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Resource": role_arn,
                    "Action": "sts:AssumeRole",
                }
            ],
        }
        name = self.random_name()
        response = self.iam.create_policy(
            PolicyName=name, PolicyDocument=json.dumps(policy_document)
        )
        self.addCleanup(
            self.iam.delete_policy, PolicyArn=response['Policy']['Arn']
        )
        return response['Policy']['Arn']

    def assert_s3_read_only_profile(self, profile_name):
        # Calls to S3 should succeed
        command = 's3api list-buckets --profile %s' % profile_name
        result = aws(command, env_vars=self.environ)
        self.assertEqual(result.rc, 0, result.stderr)

        # Calls to other services should not
        command = 'iam list-groups --profile %s' % profile_name
        result = aws(command, env_vars=self.environ)
        self.assertNotEqual(result.rc, 0, result.stdout)
        self.assertIn('AccessDenied', result.stderr)

    def test_recursive_assume_role(self):
        # Create the final role, the one that will actually have access to s3
        final_role = self.create_role(self.role_policy, S3_READ_POLICY_ARN)

        # Create the role that can assume the final role
        middle_policy_arn = self.create_assume_policy(final_role['Arn'])
        middle_role = self.create_role(self.role_policy, middle_policy_arn)

        # Create a user that can only assume the middle-man role, and then get
        # static credentials for it.
        user_policy_arn = self.create_assume_policy(middle_role['Arn'])
        user = self.create_user([user_policy_arn])
        user_creds = self.create_creds(user['UserName'])

        # Setup the config file with the profiles we'll be using. For
        # convenience static credentials are placed here instead of putting
        # them in the credentials file.
        config = (
            '[default]\n'
            'aws_access_key_id = %s\n'
            'aws_secret_access_key = %s\n'
            '[profile middle]\n'
            'source_profile = default\n'
            'role_arn = %s\n'
            '[profile final]\n'
            'source_profile = middle\n'
            'role_arn = %s\n'
        )
        config = config % (
            user_creds['AccessKeyId'],
            user_creds['SecretAccessKey'],
            middle_role['Arn'],
            final_role['Arn'],
        )
        with open(self.config_file, 'w') as f:
            f.write(config)

        # Wait for IAM permissions to propagate
        middle_creds = self.wait_for_assume_role(
            role_arn=middle_role['Arn'],
            access_key=user_creds['AccessKeyId'],
            secret_key=user_creds['SecretAccessKey'],
        )
        self.wait_for_assume_role(
            role_arn=final_role['Arn'],
            access_key=middle_creds['AccessKeyId'],
            secret_key=middle_creds['SecretAccessKey'],
            token=middle_creds['SessionToken'],
        )

        # Configure our credentials file to be THE credentials file
        self.environ['AWS_CONFIG_FILE'] = self.config_file

        self.assert_s3_read_only_profile(profile_name='final')

    def test_assume_role_with_credential_source(self):
        # Create a role with read access to S3
        role = self.create_role(self.role_policy, S3_READ_POLICY_ARN)

        # Create a user that can assume the role and get static credentials
        # for it.
        user_policy_arn = self.create_assume_policy(role['Arn'])
        user = self.create_user([user_policy_arn])
        user_creds = self.create_creds(user['UserName'])

        # Setup the config file with the profile we'll be using.
        config = (
            '[profile assume]\n'
            'role_arn = %s\n'
            'credential_source = Environment\n'
        )
        config = config % role['Arn']
        with open(self.config_file, 'w') as f:
            f.write(config)

        # Wait for IAM permissions to propagate
        self.wait_for_assume_role(
            role_arn=role['Arn'],
            access_key=user_creds['AccessKeyId'],
            secret_key=user_creds['SecretAccessKey'],
        )

        # Setup the environment so that our new config file is THE config
        # file and add the expected credentials since we're using the
        # environment as our credential source.
        self.environ['AWS_CONFIG_FILE'] = self.config_file
        self.environ['AWS_SECRET_ACCESS_KEY'] = user_creds['SecretAccessKey']
        self.environ['AWS_ACCESS_KEY_ID'] = user_creds['AccessKeyId']

        self.assert_s3_read_only_profile(profile_name='assume')



import os
import unittest

from awscli.clidocs import (
    GLOBAL_OPTIONS_FILE,
    GLOBAL_OPTIONS_SYNOPSIS_FILE,
    GlobalOptionsDocumenter,
)
from awscli.testutils import create_clidriver


class TestGlobalOptionsDocumenter(unittest.TestCase):
    def setUp(self):
        self.driver = create_clidriver()
        self.help_command = self.driver.create_help_command()
        self.globals = GlobalOptionsDocumenter(self.help_command)

    def test_doc_global_options_match_saved_content(self):
        with open(GLOBAL_OPTIONS_FILE) as f:
            self.assertEqual(self.globals.doc_global_options(), f.read())

    def test_doc_global_synopsis_match_saved_content(self):
        with open(GLOBAL_OPTIONS_SYNOPSIS_FILE) as f:
            self.assertEqual(self.globals.doc_global_synopsis(), f.read())



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.s3.fileinfo import FileInfo
from awscli.testutils import unittest


class TestIsGlacierCompatible(unittest.TestCase):
    def setUp(self):
        self.file_info = FileInfo('bucket/key')
        self.file_info.associated_response_data = {'StorageClass': 'GLACIER'}

    def test_operation_is_glacier_compatible(self):
        self.file_info.operation_name = 'delete'
        self.assertTrue(self.file_info.is_glacier_compatible())

    def test_download_operation_is_not_glacier_compatible(self):
        self.file_info.operation_name = 'download'
        self.assertFalse(self.file_info.is_glacier_compatible())

    def test_copy_operation_is_not_glacier_compatible(self):
        self.file_info.operation_name = 'copy'
        self.assertFalse(self.file_info.is_glacier_compatible())

    def test_operation_is_glacier_compatible_for_non_glacier(self):
        self.file_info.operation_name = 'download'
        self.file_info.associated_response_data = {'StorageClass': 'STANDARD'}
        self.assertTrue(self.file_info.is_glacier_compatible())

    def test_move_operation_is_not_glacier_compatible_for_s3_source(self):
        self.file_info.operation_name = 'move'
        self.file_info.src_type = 's3'
        self.assertFalse(self.file_info.is_glacier_compatible())

    def test_move_operation_is_glacier_compatible_for_local_source(self):
        self.file_info.operation_name = 'move'
        self.file_info.src_type = 'local'
        self.assertTrue(self.file_info.is_glacier_compatible())

    def test_response_is_not_glacier(self):
        self.file_info.associated_response_data = {'StorageClass': 'STANDARD'}
        self.assertTrue(self.file_info.is_glacier_compatible())

    def test_response_missing_storage_class(self):
        self.file_info.associated_response_data = {'Key': 'Foo'}
        self.assertTrue(self.file_info.is_glacier_compatible())

    def test_restored_object_is_glacier_compatible(self):
        self.file_info.operation_name = 'download'
        self.file_info.associated_response_data = {
            'StorageClass': 'GLACIER',
            'Restore': 'ongoing-request="false", expiry-date="..."',
        }
        self.assertTrue(self.file_info.is_glacier_compatible())

    def test_ongoing_restore_is_not_glacier_compatible(self):
        self.file_info.operation_name = 'download'
        self.file_info.associated_response_data = {
            'StorageClass': 'GLACIER',
            'Restore': 'ongoing-request="true", expiry-date="..."',
        }
        self.assertFalse(self.file_info.is_glacier_compatible())



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import json
import os

import pytest
from botocore.utils import generate_login_cache_key

from awscli.customizations.login.logout import LogoutCommand
from tests import mock


@pytest.fixture
def login_session_id():
    return "arn:aws:sts::012345678901:assumed-role/Login/foo"


@pytest.fixture
def config(login_session_id):
    return {
        'profiles': {
            'default': {
                'region': 'us-east-1',
            },
            'login-profile': {
                'region': 'us-east-1',
                'login_session': login_session_id,
            },
        }
    }


@pytest.fixture
def session(config):
    session = mock.Mock()
    session.full_config = config
    return session


@pytest.fixture
def login_cache_dir(tmpdir, login_session_id):
    cache_key = generate_login_cache_key(login_session_id)
    tmpdir.join(f"{cache_key}.json").write(b'')
    return tmpdir


@pytest.fixture
def valid_token_structure():
    return {
        "accessToken": {
            "accessKeyId": "ASIATESTACCESSKEY123",
            "secretAccessKey": "test-secret-key",
            "sessionToken": "test-session-token",
            "accountId": "012345678901",
            "expiresAt": "2025-11-10T12:00:00Z"
        },
        "tokenType": "Bearer",
        "clientId": "arn:aws:signin:::devtools/test",
        "refreshToken": "test-refresh-token",
        "idToken": "test-id-token",
        "dpopKey": (
            "-----BEGIN EC PRIVATE KEY-----\n"
            "test-key\n"
            "-----END EC PRIVATE KEY-----"
        )
    }


@pytest.fixture
def logout_cmd(session, login_cache_dir):
    logout = LogoutCommand(
        session=session,
        cache_dir=login_cache_dir,
    )
    return logout


class TestLogoutCommand:
    @mock.patch('awscli.customizations.login.logout.os.remove')
    def test_logout_deletes_cache_files(
        self,
        mock_remove,
        logout_cmd,
        session,
        login_cache_dir,
        login_session_id,
        capsys,
    ):
        session.profile = 'login-profile'
        cache_key = generate_login_cache_key(login_session_id)

        parsed_args = mock.Mock()
        parsed_args.all = False
        logout_cmd._run_main(parsed_args, mock.Mock())

        expected_calls = [
            mock.call(os.path.join(login_cache_dir, f"{cache_key}.json")),
        ]
        mock_remove.assert_has_calls(expected_calls)

        captured = capsys.readouterr()
        assert (
            "Removed cached login credentials for profile 'login-profile'"
            in captured.out
        )

    @mock.patch('awscli.customizations.login.logout.os.remove')
    def test_logout_no_login_session(
        self, mock_remove, logout_cmd, session, capsys
    ):
        session.profile = 'default'

        parsed_args = mock.Mock()
        parsed_args.all = False
        logout_cmd._run_main(parsed_args, mock.Mock())
        mock_remove.assert_not_called()

        captured = capsys.readouterr()
        assert (
            captured.err
            == "warning: no login session found for profile 'default'\n"
        )

    def test_logout_all_deletes_multiple_tokens(
        self, logout_cmd, login_cache_dir, valid_token_structure, capsys
    ):
        token_files = []
        for i in range(3):
            token_file = login_cache_dir.join(f"token_{i}.json")
            token_file.write(json.dumps(valid_token_structure))
            token_files.append(token_file)

        for token_file in token_files:
            assert token_file.exists()

        parsed_args = mock.Mock()
        parsed_args.all = True
        logout_cmd._run_main(parsed_args, mock.Mock())

        for token_file in token_files:
            assert not token_file.exists()

        captured = capsys.readouterr()
        assert "Removed 3 cached login credentials" in captured.out

    def test_logout_all_ignores_invalid_json(
        self, logout_cmd, login_cache_dir, valid_token_structure, capsys
    ):
        valid_token_file = login_cache_dir.join("valid_token.json")
        valid_token_file.write(json.dumps(valid_token_structure))

        invalid_json_file = login_cache_dir.join("invalid.json")
        invalid_json_file.write("{ this is not valid json }")

        assert valid_token_file.exists()
        assert invalid_json_file.exists()

        parsed_args = mock.Mock()
        parsed_args.all = True
        logout_cmd._run_main(parsed_args, mock.Mock())

        assert not valid_token_file.exists()

        assert invalid_json_file.exists()

        captured = capsys.readouterr()
        assert "Removed 1 cached login credential" in captured.out

    def test_logout_all_ignores_non_token_files(
        self, logout_cmd, login_cache_dir, valid_token_structure, capsys
    ):
        valid_token_file = login_cache_dir.join("valid_token.json")
        valid_token_file.write(json.dumps(valid_token_structure))

        non_token_json = {
            "someOtherKey": "value",
            "anotherKey": {
                "nestedData": "test"
            }
        }
        non_token_file = login_cache_dir.join("non_token.json")
        non_token_file.write(json.dumps(non_token_json))

        assert valid_token_file.exists()
        assert non_token_file.exists()

        parsed_args = mock.Mock()
        parsed_args.all = True
        logout_cmd._run_main(parsed_args, mock.Mock())

        assert not valid_token_file.exists()

        assert non_token_file.exists()

        captured = capsys.readouterr()
        assert "Removed 1 cached login credential" in captured.out

    def test_logout_all_with_no_tokens(
        self, logout_cmd, tmpdir, session, capsys
    ):
        empty_cache_dir = tmpdir.mkdir("empty_cache")

        logout_cmd_empty = LogoutCommand(
            session=session,
            cache_dir=empty_cache_dir,
        )

        parsed_args = mock.Mock()
        parsed_args.all = True
        logout_cmd_empty._run_main(parsed_args, mock.Mock())

        captured = capsys.readouterr()
        assert "No cached login session tokens found" in captured.out

    def test_logout_all_when_cache_dir_missing(
        self, session, tmpdir, capsys
    ):
        non_existent_cache_dir = tmpdir.join("non_existent_cache")

        assert not non_existent_cache_dir.exists()

        logout_cmd_missing = LogoutCommand(
            session=session,
            cache_dir=non_existent_cache_dir,
        )

        parsed_args = mock.Mock()
        parsed_args.all = True
        logout_cmd_missing._run_main(parsed_args, mock.Mock())

        captured = capsys.readouterr()
        assert "No cached login session tokens found" in captured.out

    def test_logout_all_validates_token_structure(
        self, logout_cmd, login_cache_dir, valid_token_structure, capsys
    ):
        valid_token_file = login_cache_dir.join("valid_token.json")
        valid_token_file.write(json.dumps(valid_token_structure))

        malformed_token_structure = {
            "accessToken": {
                "secretAccessKey": "test-secret-key",
                "sessionToken": "test-session-token",
                "accountId": "012345678901",
                "expiresAt": "2025-11-10T12:00:00Z"
            },
            "tokenType": "Bearer",
            "clientId": "arn:aws:signin:::devtools/test",
            "refreshToken": "test-refresh-token",
            "idToken": "test-id-token"
        }

        malformed_token_file = login_cache_dir.join("malformed_token.json")
        malformed_token_file.write(json.dumps(malformed_token_structure))

        assert valid_token_file.exists()
        assert malformed_token_file.exists()

        parsed_args = mock.Mock()
        parsed_args.all = True
        logout_cmd._run_main(parsed_args, mock.Mock())

        assert not valid_token_file.exists()

        assert malformed_token_file.exists()

        captured = capsys.readouterr()
        assert "Removed 1 cached login credential" in captured.out
